# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""MultiprocessingPipeline.
"""

# pylint: disable=missing-class-docstring

from abc import abstractmethod
import copy
import io
import logging
import multiprocessing
import concurrent.futures
from concurrent.futures import (
    FIRST_COMPLETED,
    Future,
    ThreadPoolExecutor,
    ProcessPoolExecutor,
)
from multiprocessing.connection import Connection
import queue
import time
from typing import Callable

from .exception import *
from .mp_global import get_verbosity_level


def _is_very_verbose_logging():
    return (
        get_verbosity_level() >= 2
        and logging.getLogger().getEffectiveLevel() >= logging.DEBUG
    )


PIPE_CONN_MSG_CMD_DATA = "data"
PIPE_CONN_MSG_CMD_DATA_FINAL = "data-final"
PIPE_CONN_MSG_DATA_CMDS = [PIPE_CONN_MSG_CMD_DATA, PIPE_CONN_MSG_CMD_DATA_FINAL]


class PipeConnectionMessage:
    def __init__(self, cmd: str, data: bytes = None) -> None:
        if data is None:
            data = bytes()
        if not isinstance(data, (bytes, bytearray)):
            raise InvalidFunctionArgument(f"Expecting data to be bytes or bytearray.")
        self.cmd = cmd
        self.data = data


class PipeConnectionIO(io.RawIOBase):
    """Allow FileIO-like interface with a Pipe Connection object. Not a
    fully functional RawIOBase but close enough to meet certain the needs
    of certain classes which can accept/use such.
    """

    # pylint: disable=no-self-use
    def __init__(self, c: Connection, is_write: bool) -> None:
        super().__init__()
        self.c = c
        self.is_write = is_write
        self._num_bytes = 0
        self._cached_fileno = self.c.fileno()
        self._eof = False
        if _is_very_verbose_logging():
            logging.debug(
                f"ConnectionIO.__init__: fileno={self._cached_fileno} is_write={is_write} {'Sender' if is_write else 'Receiver'}."
            )

    @staticmethod
    def create_reader_writer_pair():
        r, w = multiprocessing.connection.Pipe(duplex=False)
        return (
            PipeConnectionIO(c=r, is_write=False),
            PipeConnectionIO(c=w, is_write=True),
        )

    def reset_num_bytes(self):
        self._num_bytes = 0

    @property
    def eof(self):
        return self._eof

    @property
    def num_bytes(self):
        return self._num_bytes

    def tell(self) -> int:
        return self.num_bytes

    def fileno(self) -> int:
        return self.c.fileno()

    def send_message(self, msg: PipeConnectionMessage):
        size = -1
        try:
            if not isinstance(msg, PipeConnectionMessage):
                raise InvalidFunctionArgument(
                    f"PipeConnectionIO.send_message: msg must be PipeConnectionMessage"
                )
            if msg.cmd is None or not isinstance(msg.cmd, str):
                raise InvalidFunctionArgument(
                    f"PipeConnectionIO.send_message: msg.cmd must be str"
                )
            if msg.data is not None and isinstance(msg.data, (bytes, bytearray)):
                size = len(msg.data)
            if msg.cmd == PIPE_CONN_MSG_CMD_DATA_FINAL and self.eof:
                raise PipeConnectionAlreadyEof(
                    f"PipeConnectionIO.send_message: cannot send PIPE_CONN_MSG_CMD_DATA_FINAL, already eof."
                )
            if _is_very_verbose_logging():
                logging.debug(
                    f"PipeConnectionIO.send_message: sending: fileno={self._cached_fileno} cmd={msg.cmd} size={size}..."
                )
            self.c.send(msg)
            if msg.cmd == PIPE_CONN_MSG_CMD_DATA_FINAL:
                self._eof = True
            if size != -1:
                self._num_bytes += size
                if _is_very_verbose_logging():
                    logging.debug(
                        f"PipeConnectionIO.send_message: sent: fileno={self._cached_fileno} cmd={msg.cmd} size={size} conv_total={self._num_bytes}"
                    )
            else:
                if _is_very_verbose_logging():
                    logging.debug(
                        f"PipeConnectionIO.send_message: sent: fileno={self._cached_fileno} cmd={msg.cmd}: data as bytes not present."
                    )
        except Exception as ex:
            logging.error(
                f"PipeConnectionIO.send_message: "
                f"fileno={self._cached_fileno} num_bytes={size} "
                f"Exception: {exc_to_string(ex)}"
            )
            raise

    def recv_message(self) -> PipeConnectionMessage:
        try:
            if _is_very_verbose_logging():
                logging.debug(
                    f"PipeConnectionIO.recv_message: receiving: fileno={self._cached_fileno}..."
                )
            msg = self.c.recv()
            if not isinstance(msg, PipeConnectionMessage):
                raise InvalidPipeConnectionMessage(
                    f"PipeConnectionIO.recv_message: "
                    f"Expecting InvalidPipeConnectionMessage but got {type(msg)}"
                )
            if msg.cmd is None:
                raise InvalidPipeConnectionMessage(
                    f"PipeConnectionIO.recv_message: "
                    f"Expecting msg.cmd to be a str but got None."
                )
            if not isinstance(msg.cmd, str):
                raise InvalidPipeConnectionMessage(
                    f"PipeConnectionIO.recv_message: "
                    f"Expecting msg.cmd to be a str but got {type(msg.cmd)}"
                )
            if msg.cmd == PIPE_CONN_MSG_CMD_DATA_FINAL:
                self._eof = True
            if msg.data is not None and isinstance(msg.data, (bytes, bytearray)):
                self._num_bytes += len(msg.data)
                if _is_very_verbose_logging():
                    logging.debug(
                        f"PipeConnectionIO.recv_message: received: fileno={self._cached_fileno} cmd={msg.cmd} size={len(msg.data)} conv_total={self._num_bytes}"
                    )
            else:
                if _is_very_verbose_logging():
                    logging.debug(
                        f"PipeConnectionIO.recv_message: received: fileno={self._cached_fileno} cmd={msg.cmd}: data as bytes not present."
                    )
            return msg
        except EOFError:
            # Do not log EOFError at this point as caller
            # may not deem it to be an error.
            raise
        except Exception as ex:
            logging.error(
                f"PipeConnectionIO.recv_message: "
                f"fileno={self._cached_fileno} "
                f"Exception: {exc_to_string(ex)}"
            )
            raise

    def _write(self, cmd, buf) -> int:
        if not self.is_write:
            raise NotImplementedError(
                f"PipeConnectionIO.write: this is not a writeable side of the pipe."
            )
        if _is_very_verbose_logging():
            logging.debug(
                f"PipeConnectionIO.write: fileno={self._cached_fileno} num_bytes={len(buf)}."
            )
        msg = PipeConnectionMessage(cmd=cmd, data=buf)
        self.send_message(msg)
        return len(buf)

    def _validate_write_buf(self, buf):
        if buf is None:
            raise InvalidFunctionArgument(
                f"PipeConnectionIO.write: buf cannot be None."
            )
        if not isinstance(buf, (bytes, bytearray)):
            raise InvalidFunctionArgument(
                f"PipeConnectionIO.write: buf must be bytes or bytearray."
            )

    def write_eof(self, buf) -> int:
        if self.eof:
            raise PipeConnectionAlreadyEof(f"PipeConnectionIO.write_eof: already eof.")
        if _is_very_verbose_logging():
            logging.debug(
                f"PipeConnectionIO.write_eof: fileno={self._cached_fileno} writing EOF."
            )
        self._validate_write_buf(buf=buf)
        return self._write(PIPE_CONN_MSG_CMD_DATA_FINAL, buf)

    def write(self, buf) -> int:
        self._validate_write_buf(buf=buf)
        if len(buf) == 0:
            # Disallow writing of zero bytes until,
            # optionally, the last write which
            # indicates EOF. Some stream writers
            # expect writing zero to be a NOP.
            if _is_very_verbose_logging():
                logging.debug(
                    f"PipeConnectionIO.write: fileno={self._cached_fileno} Skipping zero-byte write."
                )
            return 0
        return self._write(PIPE_CONN_MSG_CMD_DATA, buf)

    def read(self, size: int = None) -> bytes:
        if self.is_write:
            raise NotImplementedError()
        if size is not None:
            raise InvalidPipeConnectionMessage(
                f"PipeConnectionIO.read: "
                f"Sender determines size, "
                f"cannot read with size specifications."
            )
        if self.eof:
            return bytes()
        try:
            msg = self.recv_message()
            if msg is None:
                raise InvalidPipeConnectionMessage(
                    f"PipeConnectionIO.read: NoneType message is unxpected."
                )
            if not isinstance(msg.cmd, str):
                raise InvalidPipeConnectionMessage(
                    f"PipeConnectionIO.read: msg.cmd must be str."
                )
            if msg.cmd not in PIPE_CONN_MSG_DATA_CMDS:
                raise InvalidPipeConnectionMessage(
                    f"PipeConnectionIO.read: "
                    f"Expecting one of '{PIPE_CONN_MSG_DATA_CMDS}' but got '{msg.cmd}'"
                )
            if not isinstance(msg.data, (bytes, bytearray)):
                raise InvalidPipeConnectionMessage(
                    f"PipeConnectionIO.read: "
                    f"Expecting msg.data to be bytes or bytearray but got {type(msg.data)}"
                )
            return msg.data
        except EOFError:
            if _is_very_verbose_logging():
                logging.debug(
                    f"PipeConnectionIO.read: fileno={self._cached_fileno} EOFError."
                )
            # Sender closes while we wait.
            return bytes()
        except Exception as ex:
            logging.error(
                f"PipeConnectionIO.read: fileno={self._cached_fileno} "
                f"Exception: {exc_to_string(ex)}"
            )
            raise

    def close(self) -> None:
        if _is_very_verbose_logging():
            logging.debug(f"PipeConnectionIO.close: fileno={self._cached_fileno}.")
            self.c.close()

    @property
    def closed(self) -> bool:
        return self.c.closed

    def seekable(self) -> bool:
        return False

    def seek(self, __offset: int, __whence: int = ...) -> int:
        raise NotImplementedError()


class PipelineWorkItem:
    def __init__(
        self,
        user_obj: object = None,
        auto_copy_attr: bool = True,
        **kwargs,
    ) -> None:
        """Create a pipeline work item.

        user_obj is a caller object that must be pickle'able and
        is always picked to/from subprocesses back to this instance.

        If auto_copy_attr is True (default), copy all attributes
        not specified OUR_ATTRIBUTES to this instance after they
        are pickle'ed back from subprocesses. Again, these must
        be pickle'able.

        kwargs is anything else you want passed to all stages. It
        is recommended you avoid kwargs and use members of this
        instance.
        """
        #
        # IMPORTANT: Update OUR_ATTRIBUTES (above) if needed.
        #
        self._cur_stage = 0
        self.user_obj = user_obj
        self.user_kwargs = kwargs
        self.exceptions: list[Exception] = None
        self.pipe_conn: Connection = None
        self._auto_copy_attr = auto_copy_attr

    #
    # Update this if you add/remove or change names of
    # attributes that should not be auto-copy'ed.
    #
    OUR_ATTRIBUTES = [
        "_cur_stage",
        "user_obj",
        "user_kwargs",
        "exceptions",
        "pipe_conn",
        "_auto_copy_attr",
    ]

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: next_stage={self._cur_stage} user_obj={self.user_obj} exceps={self.exceptions}"

    @property
    def auto_copy_attr(self):
        return self._auto_copy_attr

    @auto_copy_attr.setter
    def auto_copy_attr(self, value):
        self._auto_copy_attr = value

    @property
    def is_failed(self):
        has_exceptions = self.exceptions is not None and len(self.exceptions) > 0
        return has_exceptions

    def increment_stage(self):
        self._cur_stage += 1

    @property
    def cur_stage(self):
        """Current stage which is usually the next stage to
        actually run, where the number increments after
        submitting the work.
        """
        return self._cur_stage

    @cur_stage.setter
    def cur_stage(self, value):
        self._cur_stage = value

    def append_exception(self, ex: Exception):
        if self.exceptions is None:
            self.exceptions = list[Exception]()
        self.exceptions.append(ex)

    def stage_complete(
        self,
        stage_num: int,  # pylint: disable=unused-argument
        wi: "PipelineWorkItem",  # pylint: disable=unused-argument
        ex: Exception,
    ):
        """If ex is None, wi is the PipelineWorkItem from the stage's
        Future result. If ex is not None, wi is this instance for use
        as needed. The stage_num is the stage number that completed.
        Overrides should either call this method via super() or provide
        equal functionality to capture any error information. Beyond
        that, use wi to capture any info from a successful result as
        desired.
        """
        if ex is not None:
            self.append_exception(ex)
        if wi.is_failed:
            if self.exceptions is None:
                self.exceptions = list[Exception]()
            self.exceptions.extend(wi.exceptions)
        self.user_obj = wi.user_obj  # Always copied.
        if self._auto_copy_attr:
            # By default, copy all of user's additions.
            # User can disable as desired.
            for k, v in wi.__dict__.items():
                if k not in PipelineWorkItem.OUR_ATTRIBUTES:
                    self.__dict__[k] = v


@dataclass(eq=True, frozen=True)
class _WorkItemStageRunCtx:
    cur_stage: int
    fut: Future
    wi: PipelineWorkItem

    def __str__(self) -> str:
        return f"cur_stage={self.cur_stage} wi={str(self.wi)} fut={str(self.fut)}"


class PipelineStage:
    def __init__(
        self,
        fn_determiner: Callable[[PipelineWorkItem], bool] = None,
        fn_worker: Callable[..., PipelineWorkItem] = None,
        **stage_kwargs,
    ) -> None:
        self.fn_determiner = fn_determiner
        self.fn_worker = fn_worker
        self.stage_kwargs = stage_kwargs

    @property
    @abstractmethod
    def is_subprocess(self):
        raise NotImplementedError(
            f"is_subprocess is not implemented. "
            f"You probably want to use either "
            f"SubprocessPipelineStage or ThreadPipelineStage"
        )

    @property
    def is_pipe_with_next_stage(self):
        return False

    def is_for_stage(self, pwi: PipelineWorkItem) -> bool:
        if self.fn_determiner is None:
            raise InvalidStateError(
                f"PipelineStage fn_determiner is None, cannot determine anything."
            )
        return self.fn_determiner(pwi)

    def perform_stage_work(self, pwi: PipelineWorkItem, **kwargs):
        if self.fn_worker is None:
            raise InvalidStateError(
                f"PipelineStage fn_worker is None, cannot work, looks like a holiday today."
            )
        result = self.fn_worker(pwi, **kwargs)
        return result


class SubprocessPipelineStage(PipelineStage):
    @property
    def is_subprocess(self):
        return True


class ThreadPipelineStage(PipelineStage):
    @property
    def is_subprocess(self):
        return False


class MultiprocessingPipeline:
    def __init__(
        self,
        stages: list[PipelineStage] = None,
        max_simultaneous_work_items=None,
        name="unnamed",
        process_initfunc=None,
        process_initargs=(),
    ) -> None:
        if stages is None:
            stages = list[PipelineStage]()
        if not isinstance(stages, list):
            raise InvalidFunctionArgument(
                f"Expecting stages to be a non-zero length list."
            )
        self._stages = stages
        self._max_simultaneous_work_items = max_simultaneous_work_items
        self._is_shutdown = False  # Accessed by parent process only.
        self._wi_to_wifut = {}
        self._wi_to_fut_cond = multiprocessing.Condition()
        self._thread_exec = ThreadPoolExecutor(thread_name_prefix=f"MpPipeline-{name}")
        self._pl_worker_future = None
        self._input_queue_lock = multiprocessing.Lock()
        self._input_queue_fut: Future = None
        self._pl_input_queue = queue.Queue()  # Accessed in parent only.
        max_workers_each = None
        if self._max_simultaneous_work_items:
            max_workers_each = int(self._max_simultaneous_work_items + 2)
        self._process_exec = ProcessPoolExecutor(
            max_workers=max_workers_each,
            initializer=process_initfunc,
            initargs=process_initargs,
        )
        self._process_exec2 = ProcessPoolExecutor(
            max_workers=max_workers_each,
            initializer=process_initfunc,
            initargs=process_initargs,
        )
        # Given Future, find work item.
        self._plfut_to_wi = dict[Future, PipelineWorkItem]()
        # Given work item, find contexts for running work, if any.
        self._running_wi_contexts = dict[PipelineWorkItem, list[_WorkItemStageRunCtx]]()
        self._fut_to_pipe_conn = dict[Future, Connection]()
        self.was_graceful_shutdown = False
        self.anomalies = list[Anomaly]()

    def add_stage(self, stage: PipelineStage):
        self._stages.append(stage)

    @property
    def num_stages(self) -> int:
        return len(self._stages)

    def _start(self):
        if self._is_shutdown:
            raise InvalidFunctionArgument(
                f"QueuedSubprocessPipeline has already been started and stopped."
            )
        if self._pl_worker_future is not None:
            return
        self._pl_worker_future = self._thread_exec.submit(
            MultiprocessingPipeline._pl_worker, self
        )

    def shutdown(self):
        if self._pl_worker_future is None:
            return
        while len(self._wi_to_wifut) > 0:
            with self._wi_to_fut_cond:
                if len(self._wi_to_wifut) == 0:
                    break
                self._wi_to_fut_cond.wait()
        self._internal_submit(None)
        self._pl_worker_future.result()
        self._pl_worker_future = None
        if self._process_exec is not None:
            self._process_exec.shutdown()
        if self._thread_exec is not None:
            self._thread_exec.shutdown()

    def _fail_all_pending(self, ex: Exception):
        wi: PipelineWorkItem
        for wi, wifut in self._wi_to_wifut.items():
            if wifut.done():
                continue
            wi.append_exception(ex)
            wifut.set_exception(ex)

    def _get_running_wi_count(self):
        return len(self._running_wi_contexts)

    def _is_wi_still_running(
        self,
        wi: PipelineWorkItem,
    ):
        wi_contexts = self._running_wi_contexts.get(wi)
        if wi_contexts is None:
            return False
        return len(wi_contexts) > 0

    def _track_running_pipeline_work(
        self,
        cur_stage: int,
        fut: Future,
        wi: PipelineWorkItem,
    ):
        """Track a work item that is running within a stage.
        Currently this is most often one per work item, but
        can be two for dual-stage pipe connection case.
        The _get_completed_work_item_info untracks what
        this tracks.
        """
        if self._plfut_to_wi.get(fut) is not None:
            raise InvalidStateError(
                f"Expected _plfut_to_wi not to already have Future."
            )
        self._plfut_to_wi[fut] = wi
        if self._running_wi_contexts.get(wi) is None:
            self._running_wi_contexts[wi] = list[_WorkItemStageRunCtx]()
        self._running_wi_contexts[wi].append(
            _WorkItemStageRunCtx(
                cur_stage=cur_stage,
                fut=fut,
                wi=wi,
            )
        )

    def _get_completed_work_item_info(
        self,
        fut: Future,
    ) -> tuple[PipelineWorkItem, _WorkItemStageRunCtx]:
        """Given a Future, untrack and return its information.
        This method untracks what _track_running_pipeline_work
        tracks.
        """
        #
        # Find/remove the Future-to-work item tracking.
        #
        wi = self._plfut_to_wi.get(fut)
        if wi is not None:
            del self._plfut_to_wi[fut]

        #
        # Given the work item, find the contexts for each
        # running stage (often just 1, could be 2 in dual
        # stage pipe conn case).
        #
        wi_contexts = self._running_wi_contexts.get(wi)
        if wi_contexts is None:
            raise InvalidStateError(
                f"Unexpected, found work item from the Future "
                f"but not its _WorkItemStageRunCtx"
            )

        #
        # Remove the context for the completed portion of work.
        #
        c = None
        for c in list(wi_contexts):
            if c.fut == fut:
                wi_contexts.remove(c)
                break
        if len(wi_contexts) == 0:
            del self._running_wi_contexts[wi]
        if c is None:
            raise InvalidStateError(
                f"Unexpected, found work item from the Future, "
                f"and its context set, but not its specific "
                f" _WorkItemStageRunCtx instance."
            )

        return (
            wi,
            c,
        )

    def _get_queued_wi(self):
        """Returns a tuple (wi, is_shutdown), where wi will
        be a work item if there's work that should run, else
        it will be None. If is_shutdown==True, caller should
        exit.
        """
        with self._input_queue_lock:
            if self._pl_input_queue.qsize() == 0:
                return (None, False)
            try:
                wi = self._pl_input_queue.get()
                return (wi, wi is None)
            except queue.Empty as ex:
                self.anomalies.append(
                    Anomaly(
                        kind=ANOMALY_KIND_EXCEPTION,
                        exception=ex,
                        message="Unexpected Empty on _pl_input_queue.get()",
                    )
                )
                return (
                    None,
                    None,
                    False,
                )

    def _handle_completed_fut(
        self,
        done_fut: Future,
    ) -> tuple[PipelineWorkItem, _WorkItemStageRunCtx, bool]:
        """Processes done_fut, possibly returning a work item.
        Returns a tuple (wi, is_shutdown), where wi is the
        work item and is_shutdown is True if caller should exit.

        The work item is either already runing work which just
        completed a stage, or new work from the input queue.
        """
        wi: PipelineWorkItem = None

        if done_fut in self._plfut_to_wi:
            #
            # Completed future pertains to already running work.
            # Update the related stage with the results. If there
            # are parallel stages for the same work item, they are
            # completed separately from this call.
            #
            self._log_state(
                ctx_str="Future completion", futs_of_interest=set([done_fut])
            )
            wi, ctx = self._get_completed_work_item_info(fut=done_fut)
            if done_fut.exception() is not None:
                #
                # Some error occurred.
                #
                wi.stage_complete(
                    stage_num=ctx.cur_stage,
                    wi=wi,
                    ex=done_fut.exception(),
                )
            else:
                #
                # A successful result.
                #
                wi_from_sp: PipelineWorkItem = done_fut.result()
                if not isinstance(wi_from_sp, PipelineWorkItem):
                    ex = PipelineResultIsNotPipelineWorkItem(
                        f"The pipeline stage was successful but it returned "
                        f"something other than a PipelineWorkItem."
                    )
                    #
                    # Error: caller did not return a work item.
                    #
                    wi.stage_complete(
                        stage_num=ctx.cur_stage,
                        wi=wi,
                        ex=ex,
                    )
                else:
                    #
                    # Stage successful for work item.
                    #
                    wi.stage_complete(
                        stage_num=ctx.cur_stage,
                        wi=wi_from_sp,
                        ex=None,
                    )
            return (wi, False)

        #
        # Future is for input queue.
        #
        if done_fut == self._input_queue_fut:
            return self._get_queued_wi()

        #
        # Future is unknown.
        #
        msg = (
            f"Expected Future for pending pipeline activity or input queue, "
            f"but the Future matched neither. {str(done_fut)}"
        )
        self.anomalies.append(
            Anomaly(
                kind=ANOMALY_KIND_EXCEPTION,
                exception=InvalidStateError(msg),
                message=msg,
            )
        )
        return self._get_queued_wi()

    def _cleanup_pipe_connections(self, done_fut: Future):
        conn = self._fut_to_pipe_conn.get(done_fut)
        if conn is not None:
            conn.close()
            del self._fut_to_pipe_conn[done_fut]

    def _handle_stages_for_wi(self, wi: PipelineWorkItem):
        if wi.cur_stage < 0 or wi.cur_stage > self.num_stages:
            wi.append_exception(
                InvalidStateError(f"Invalid pipeline stage next_stage={wi.cur_stage}")
            )
            wi.cur_stage = self.num_stages
            self._set_wi_to_finished(wi)
            return

        #
        # Advance to next stage, submit work item.
        #
        while wi.cur_stage < self.num_stages:
            next_stage = self._stages[wi.cur_stage]
            if next_stage.is_pipe_with_next_stage:
                if self._try_submit_to_dual_stage_with_pipe(wi=wi):
                    # Work item is now running on two stages w/pipe.
                    break
            elif self._try_submit_to_stage(wi=wi):
                # Work item is now running on one stage.
                break
            if wi.is_failed:
                # Failure occurred trying to submit.
                break
            # Work item was not submitted, not failed.
            # Ask next stage if it wants the work item.
            wi.increment_stage()
        if not self._is_wi_still_running(wi=wi) and wi.cur_stage >= self.num_stages:
            self._set_wi_to_finished(wi)

    def _pl_worker(self):
        try:
            is_shutdown: bool = False
            while not is_shutdown:

                #
                # Check for new work.
                #
                wi, is_shutdown = self._get_queued_wi()
                if is_shutdown:
                    break

                if wi is not None:
                    #
                    # New work available.
                    #
                    wi_needing_attention = set([wi])
                else:
                    #
                    # Wait on all pending work plus
                    # the new work input queue.
                    #
                    all_futs = set(self._plfut_to_wi)
                    self._renew_input_queue_future()
                    all_futs.add(self._input_queue_fut)
                    done_futs, _ = concurrent.futures.wait(
                        fs=all_futs, return_when=FIRST_COMPLETED
                    )

                    #
                    # Something completed, gather list.
                    #
                    wi_needing_attention = set()
                    for done_fut in done_futs:
                        self._cleanup_pipe_connections(done_fut=done_fut)
                        wi, is_sd = self._handle_completed_fut(done_fut=done_fut)
                        is_shutdown = is_shutdown or is_sd
                        if wi is not None:
                            wi_needing_attention.add(wi)
                    if is_shutdown:
                        break

                #
                # Check completed results...
                #
                for wi in wi_needing_attention:

                    is_wi_still_running = self._is_wi_still_running(wi=wi)
                    if is_wi_still_running:
                        # Possible states:
                        #   * Two stages running, one just succeeded.
                        #   * Two stages running, one just failed.
                        # In either case, either the earlier or later stage is
                        # still running. Even if this is the later stage, the
                        # pipeline waits until the "paired earlier" stage completes.
                        # When a stage completes and is_wi_still_running is False,
                        # that is the time run the next wi.cur_stage stage.
                        # Wait until the remaining stage completes.
                        continue

                    if wi.is_failed:
                        # Failed and nothing else running for work item.
                        self._set_wi_to_finished(wi)
                        continue

                    self._handle_stages_for_wi(wi=wi)
        except Exception as ex:
            self._fail_all_pending(ex)
            raise
        self.was_graceful_shutdown = True

    def _submit_to_stage(
        self,
        stage: PipelineStage,
        wi: PipelineWorkItem,
        pipe_conn: Connection = None,
        use_second_pool: bool = False,
    ) -> Future:

        # Copy so work item state for submission.
        stage_wi_copy = copy.copy(wi)
        stage_wi_copy.pipe_conn = pipe_conn

        #
        # Submit the work item to the stage.
        #
        kwargs = stage.stage_kwargs
        if stage_wi_copy.user_kwargs is not None:
            kwargs = kwargs | stage_wi_copy.user_kwargs
        if stage.is_subprocess:
            if not use_second_pool:
                fut = self._process_exec.submit(
                    stage.perform_stage_work, stage_wi_copy, **kwargs
                )
            else:
                fut = self._process_exec2.submit(
                    stage.perform_stage_work, stage_wi_copy, **kwargs
                )
        else:
            fut = self._thread_exec.submit(
                stage.perform_stage_work, stage_wi_copy, **kwargs
            )

        #
        # Track the submitted work item.
        #
        self._track_running_pipeline_work(
            cur_stage=wi.cur_stage,
            fut=fut,
            wi=wi,
        )
        return fut

    def _is_stage_for_work_item(
        self,
        wi: PipelineWorkItem,
        stage_num: int = None,
    ) -> bool:
        """Returns True if next stage wants work item, else False."""
        if stage_num is None:
            stage_num = wi.cur_stage
        next_stage = self._stages[stage_num]
        try:
            # Ask stage if there's interest in this work item.
            return next_stage.is_for_stage(wi)
        except Exception as ex:
            # Unexpected failure, move past end of last stage
            # to trigger failure completion by caller.
            wi.append_exception(ex)
            return False

    def _try_submit_to_stage(self, wi: PipelineWorkItem) -> bool:
        if not self._is_stage_for_work_item(wi=wi):
            if wi.is_failed:
                wi.cur_stage = self.num_stages
            return False
        fut = self._submit_to_stage(stage=self._stages[wi.cur_stage], wi=wi)
        self._log_state(ctx_str="single stage submission", futs_of_interest=set([fut]))
        wi.increment_stage()
        return True

    def _try_submit_to_dual_stage_with_pipe(
        self,
        wi: PipelineWorkItem,
    ) -> bool:
        """Some stages request to run the work item in two stages
        at once, the current and next stage, with a pipe to send
        data from current to next stage.
        """
        #
        # Last stage cannot request dual stage run.
        #
        if wi.cur_stage == self.num_stages - 1:
            raise PipelineLastStageError(
                f"Last stage of pipeline cannot request pipe, there is no next stage."
            )

        #
        # Current and next stage must approve.
        #
        if not self._is_stage_for_work_item(wi=wi, stage_num=wi.cur_stage):
            if wi.is_failed:
                wi.cur_stage = self.num_stages
            return False
        if not self._is_stage_for_work_item(wi=wi, stage_num=wi.cur_stage + 1):
            if wi.is_failed:
                wi.cur_stage = self.num_stages
            return False

        #
        # Create the pipe connections.
        #
        conn_w: Connection = None
        conn_r: Connection = None
        fut_w: Future = None  # current stage
        fut_r: Future = None  # next stage if current stage requests pipe.
        conn_r, conn_w = multiprocessing.Pipe(duplex=False)

        #
        # Submit work item to writer stage.
        #
        next_stage = self._stages[wi.cur_stage]
        fut_w = self._submit_to_stage(stage=next_stage, wi=wi, pipe_conn=conn_w)

        #
        # Submit work item to reader stage.
        #
        wi.increment_stage()
        next_stage = self._stages[wi.cur_stage]
        fut_r = self._submit_to_stage(stage=next_stage, wi=wi, pipe_conn=conn_r)
        self._log_state(
            ctx_str="dual stage submission", futs_of_interest=set([fut_w, fut_r])
        )

        wi.increment_stage()

        self._fut_to_pipe_conn[fut_w] = conn_w
        self._fut_to_pipe_conn[fut_r] = conn_r
        return True

    def _set_wi_to_finished(
        self,
        wi: PipelineWorkItem,
    ):
        fut: Future
        if self._is_wi_still_running(wi=wi):
            self._log_state(
                ctx_str="_set_wi_to_finished: wi still running", only_of_interest=False
            )
            raise InvalidStateError(f"Work item has ongoing work being tracked.")
        with self._wi_to_fut_cond:
            fut = self._wi_to_wifut.get(wi)
            if fut is None:
                raise InvalidStateError(f"Cannot find Future for work item.")
            del self._wi_to_wifut[wi]
            if _is_very_verbose_logging():
                self._log_state(
                    ctx_str="fully completed user future",
                    futs_of_interest=set([fut]),
                    only_of_interest=False,
                )
            self._wi_to_fut_cond.notify_all()
            fut.set_result(wi)

    def _renew_input_queue_future(self):
        with self._input_queue_lock:
            self._input_queue_fut = Future()
            qsize = self._pl_input_queue.qsize()
            if qsize > 0:
                self._input_queue_fut.set_result(qsize)

    def _internal_submit(
        self,
        work_item: PipelineWorkItem,
    ) -> Future:

        self._start()

        # Track until all stages complete for work item.
        wi_fut = Future()
        self._wi_to_wifut[work_item] = wi_fut

        self._pl_input_queue.put(work_item)
        with self._input_queue_lock:
            if self._pl_input_queue.qsize() > 0:
                if (
                    self._input_queue_fut is not None
                    and not self._input_queue_fut.done()
                ):
                    self._input_queue_fut.set_result(self._pl_input_queue.qsize())
        return wi_fut

    def _log_state(
        self,
        ctx_str=None,
        futs_of_interest: set[Future] = None,
        only_of_interest: bool = True,
    ):
        if not _is_very_verbose_logging():
            return
        if futs_of_interest is None:
            futs_of_interest = set()
        if ctx_str is not None:
            logging.debug(ctx_str)

        header = False
        for wi, wifut in self._wi_to_wifut.items():
            marker = ""
            if wifut in futs_of_interest:
                marker = " <---"
            if not only_of_interest or wifut in futs_of_interest:
                if not header:
                    logging.debug(f"  top-level of user:")
                    header = True
                logging.debug(f"    fut={str(wifut)} {str(wi)}{marker}")

        header = False
        for wi, plctx_list in self._running_wi_contexts.items():
            for plctx in plctx_list:
                marker = ""
                if plctx.fut in futs_of_interest:
                    marker = " <---"
                if not only_of_interest or plctx.fut in futs_of_interest:
                    if not header:
                        logging.debug(f"  pipeline stage work:")
                        logging.debug(f"    {str(wi)}")
                        header = True
                    logging.debug(f"      {str(plctx)}{marker}")

    def submit(
        self,
        work_item: PipelineWorkItem,
    ) -> Future:
        if work_item is None:
            raise InvalidFunctionArgument(f"work_item must be a PipelineWorkItem.")
        if self._wi_to_wifut.get(work_item) is not None:
            raise InvalidFunctionArgument(f"work_item is already in the pipeline.")
        return self._internal_submit(
            work_item=work_item,
        )



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Verify backups.
"""
# pylint: disable=missing-class-docstring
import gzip
import io
import logging
import os
from pathlib import Path
from concurrent import futures
from concurrent.futures import FIRST_COMPLETED
import shutil
import tempfile

from atbu.common.util_helpers import clear_file

from ..common.exception import (
    CompareBytesMistmatchError,
    InvalidStateError,
    VerifyFailure,
    VerifyFilePathNotFoundError,
    exc_to_string,
    AlreadyUsedError,
)
from ..common.hasher import (
    GlobalHasherDefinitions,
)
from .backup_core import (
    BACKUP_OPERATION_NAME_VERIFY,
    DEFAULT_MAX_SIMULTANEOUS_FILE_BACKUPS,
    BackupAnomaly,
    log_anomalies_report,
    StorageDefinition,
    BackupFileInformation,
    StorageFileRetriever,
    file_operation_futures_to_results,
    BackupPipelineWorkItem,
    is_qualified_for_operation,
    run_operation_stage,
)
from .backup_selections import (
    SpecificBackupSelection,
    get_all_specific_backup_file_info,
    user_specifiers_to_selections,
    verify_specific_backup_selection_list,
)
from ..common.mp_global import (
    get_process_pool_exec_init_func,
    get_process_pool_exec_init_args,
)
from ..common.mp_pipeline import (
    MultiprocessingPipeline,
    SubprocessPipelineStage,
)


class VerifyFile(StorageFileRetriever):
    def __init__(
        self,
        storage_def: StorageDefinition,
        file_info: BackupFileInformation,
        is_perform_local_compare: bool,
        local_compare_root_location: str,
        temp_dir: str,
    ):
        super().__init__(file_info=file_info, storage_def=storage_def)
        #
        # No local compare: local_compare=False local_compare_root_location=N/A
        # Local compare against orig location: local_compare=True local_compare_root_location=None
        # Local compare against specified location: local_compare=True local_compare_root_location=<specified_location>
        #
        self._temp_dir = temp_dir
        self.is_perform_local_compare = is_perform_local_compare
        self.local_compare_root_location = None
        self.local_compare_path = None
        self.local_compare_path_str = None
        if self.is_perform_local_compare:
            if local_compare_root_location:
                # Some location off a specified root location.
                self.local_compare_root_location = Path(local_compare_root_location)
                self.local_compare_path = (
                    self.local_compare_root_location
                    / self.file_info.path_without_discovery_path
                )
                self.local_compare_path_str = str(self.local_compare_path)
            else:
                # Original backup path.
                self.local_compare_path = Path(self.file_info.path)
                self.local_compare_path_str = self.file_info.path
        self.local_compare_file: io.FileIO = None
        self.total_compare_bytes: int = 0
        self._temp_dest_path: Path = None
        self._temp_dest_file: io.FileIO = None

    @property
    def path_for_logging(self) -> str:
        if self.local_compare_path_str is not None:
            return self.local_compare_path_str
        return super().path_for_logging

    def prepare_destination(self):
        if self.is_perform_local_compare:
            if not self.local_compare_path.exists():
                raise VerifyFilePathNotFoundError(
                    f"The verify file destination path was not found: {str(self.local_compare_path)}"
                )
            self.local_compare_file = open(self.local_compare_path_str, "rb")

        if self.is_compressed:
            #
            # Decomrpession case: decompress to temp file.
            #
            dest_file_fd, self._temp_dest_path = tempfile.mkstemp(
                prefix="atbu_z_", dir=self._temp_dir, text=False
            )
            self._temp_dest_path = Path(self._temp_dest_path)
            self._temp_dest_file = io.FileIO(
                file=dest_file_fd, mode="w+b", closefd=True
            )
            # Disable default cleartext hashing because completion code will
            # perform a cleartext hashing on the final decompressed file.
            self.disable_cleartext_hashing()
            return

    def attempt_failed_cleanup(self):
        if self.local_compare_file:
            self.local_compare_file.close()
            self.local_compare_file = None

    def report_progress(self, percent: int):
        logging.info(f"{percent: >3}% completed of {self.file_info.path_without_root}")

    def process_decrypted_chunk(self, decrypted_chunk: bytes):
        if self.is_compressed:
            self._temp_dest_file.write(decrypted_chunk)
            return
        if not self.is_perform_local_compare:
            return
        if not self.local_compare_file:
            raise InvalidStateError(
                f"VerifyFile: Expected self.dest_file to be set/open."
            )
        cur_pos = self.local_compare_file.tell()
        local_chunk = self.local_compare_file.read(len(decrypted_chunk))
        if decrypted_chunk != local_chunk:
            raise VerifyFailure(
                f"VerifyFile: The file verification failed: "
                f"cur_pos={cur_pos} "
                f"storage_chunk_size={len(decrypted_chunk)} "
                f"local_chunk_size={len(local_chunk)} "
                f"path={self.local_compare_path_str}"
            )
        self.total_compare_bytes += len(decrypted_chunk)

    def download_completed(self):
        if not self.is_compressed:
            if self._temp_dest_file is not None:
                raise InvalidStateError(
                    f"_temp_dest_file should not be set if not compression."
                )
        else:
            self._temp_dest_file.seek(0, io.SEEK_SET)
            self.total_cleartext_bytes = (
                0  # zero-out compressed size, calc decomp size.
            )
            cleartext_hasher = (
                GlobalHasherDefinitions().create_hasher()
            )  # calc decomp hash.
            with gzip.GzipFile(
                mode="rb",
                fileobj=self._temp_dest_file,
            ) as input_file_gz:
                while True:
                    decrypted_decomp_chunk = input_file_gz.read(1024 * 1024 * 25)
                    if len(decrypted_decomp_chunk) == 0:
                        break
                    self.total_cleartext_bytes += len(decrypted_decomp_chunk)
                    cleartext_hasher.update_all(decrypted_decomp_chunk)
                    if self.local_compare_file is not None:
                        cur_pos = self.local_compare_file.tell()
                        local_chunk = self.local_compare_file.read(
                            len(decrypted_decomp_chunk)
                        )
                        self.total_compare_bytes += len(decrypted_decomp_chunk)
                        if decrypted_decomp_chunk != local_chunk:
                            raise VerifyFailure(
                                f"VerifyFile: The file verification failed: "
                                f"cur_pos={cur_pos} "
                                f"storage_chunk_size={len(decrypted_decomp_chunk)} "
                                f"local_chunk_size={len(local_chunk)} "
                                f"path={self.local_compare_path_str}"
                            )
            self._cleartext_digest = cleartext_hasher.get_primary_hexdigest()
            # Verify of a compressed file decrypts the compressed file to a
            # subdirectory in the user's temp directory, where it is then
            # decompressed for verify of the decompressed bytes. Zero'ing the
            # file is a best-effort and should not be considered a secure
            # clear. Python states it acquires temp file with exclusive access,
            # and it is kept open through the following.
            clear_file(fileobj=self._temp_dest_file, byte_pattern=[0x00])
            self._temp_dest_file.close()
            os.remove(self._temp_dest_path)

        if self.local_compare_file:
            self.local_compare_file.close()
            self.local_compare_file = None

        self.perform_common_checks(
            log_msg_prefix_str="VerifyFile",
            file_path_for_logging=self.path_for_logging,
            orig_file_info=self.file_info,
        )

        if (
            self.is_perform_local_compare
            and self.total_compare_bytes != self.file_info.size_in_bytes
        ):
            raise CompareBytesMistmatchError(
                f"VerifyFile: Bytes compared with local file are mismatched to file size: "
                f"compare_bytes={self.total_compare_bytes} local_bytes={self.file_info.size_in_bytes}"
            )

    def final_cleanup(self):
        if self.local_compare_file:
            self.local_compare_file.close()
            self.local_compare_file = None


class Verify:
    MAX_SIMULTANEOUS_FILES = DEFAULT_MAX_SIMULTANEOUS_FILE_BACKUPS

    def __init__(
        self,
        verify_selections: list[SpecificBackupSelection],
        local_compare: bool = False,
        local_compare_root_location: str = None,
    ) -> None:
        if len(verify_selections) == 0:
            raise InvalidStateError(
                f"VerifyFile requires sbs_list to have at least one SpecificBackupSelection instance."
            )
        self._temp_dir = tempfile.mkdtemp(prefix="atbu_vertmp_")
        self.verify_selections = verify_selections
        if not isinstance(self.verify_selections[0].storage_def, StorageDefinition):
            raise ValueError(f"The storage_def must be a StorageDefinition.")
        self._storage_def = self.verify_selections[0].storage_def
        self.selected_files: list[BackupFileInformation] = None
        self.local_compare = local_compare
        self.local_compare_root_location = local_compare_root_location
        self._subprocess_pipeline = MultiprocessingPipeline(
            name="Verify",
            max_simultaneous_work_items=min(os.cpu_count() // 2, 15),
            stages=[
                SubprocessPipelineStage(
                    fn_determiner=is_qualified_for_operation,
                    fn_worker=run_operation_stage,
                )
            ],
            process_initfunc=get_process_pool_exec_init_func(),
            process_initargs=get_process_pool_exec_init_args(),
        )
        self.anomalies: list[BackupAnomaly] = []
        self.success_count = 0
        self._is_used = False

    def shutdown(self):
        if self._subprocess_pipeline is not None:
            self._subprocess_pipeline.shutdown()
        try:

            def on_error(_, path, exc_info):
                value = ""
                if (
                    exc_info is not None
                    and len(exc_info) >= 3
                    and exc_info[2] is not None
                ):
                    value = f" ({exc_info[2]})"
                logging.warning(f"Failed to delete temp file: {path}{value}")

            if (
                self._temp_dir is not None
                and len(self._temp_dir) > 0
                and os.path.isdir(self._temp_dir)
            ):
                shutil.rmtree(
                    path=self._temp_dir, ignore_errors=False, onerror=on_error
                )
        except Exception as ex:
            logging.error(
                f"Unhandled exception while cleaning up the verify temp folder: {self._temp_dir} ex={ex}"
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()
        return False

    @property
    def storage_def(self) -> StorageDefinition:
        return self._storage_def

    @property
    def is_failed(self):
        return len(self.anomalies) > 0

    def _verify_files(self):
        if self._is_used:
            raise AlreadyUsedError(f"This instance has already been used.")
        self._is_used = True
        logging.info(f"Starting verify from '{self.storage_def.storage_def_name}'...")
        pending_verifications = set()
        try:
            logging.info(f"Scheduling verification jobs...")
            for file_info in self.selected_files:
                if file_info.is_unchanged_since_last:
                    logging.debug(
                        f"unchanged: path={file_info.path} "
                        f"path_wo_root={file_info.path_without_root} "
                        f"backing={file_info.backing_fi.path}"
                    )
                else:
                    logging.debug(
                        f"changed: path={file_info.path} "
                        f"path_wo_root={file_info.path_without_root} "
                        f"backing={file_info.backing_fi is not None}"
                    )
                logging.debug(f"")
                verify_file = VerifyFile(
                    storage_def=self.storage_def,
                    file_info=file_info,
                    is_perform_local_compare=self.local_compare,
                    local_compare_root_location=self.local_compare_root_location,
                    temp_dir=self._temp_dir,
                )
                future = self._subprocess_pipeline.submit(
                    work_item=BackupPipelineWorkItem(
                        operation_name=BACKUP_OPERATION_NAME_VERIFY,
                        file_info=file_info,
                        is_qualified=True,
                        operation_runner=verify_file,
                    )
                )
                pending_verifications.add(future)
        except Exception as ex:
            logging.error(
                f"Unexpected exception during _verify_files. {exc_to_string(ex)}"
            )
            raise
        finally:
            logging.info(f"Wait for verify file operations to complete...")
            while len(pending_verifications) > 0:
                done, pending_verifications = futures.wait(
                    fs=pending_verifications, return_when=FIRST_COMPLETED
                )
                num_completed = len(done)
                fi_complete = []
                new_anomalies: list[BackupAnomaly] = []
                file_operation_futures_to_results(
                    fs=done,
                    fi_list=fi_complete,
                    anomalies=new_anomalies,
                    the_operation="Verify",
                )
                self.success_count += num_completed - len(new_anomalies)
                self.anomalies.extend(new_anomalies)
                self.anomalies.extend(self._subprocess_pipeline.anomalies)
            self._subprocess_pipeline.shutdown()

    def verify_files(self):
        self.selected_files = get_all_specific_backup_file_info(
            sbs_list=self.verify_selections
        )
        self._verify_files()
        logging.info(f"All file verify operations have completed.")
        if len(self.anomalies) == 0:
            logging.info(f"***************")
            logging.info(f"*** SUCCESS ***")
            logging.info(f"***************")
            logging.info(f"No errors detected during verify.")
        else:
            log_anomalies_report(anomalies=self.anomalies)
        logging.info(f"{'Total files ':.<45} {len(self.selected_files)}")
        logging.info(f"{'Total errors ':.<45} {len(self.anomalies)}")
        logging.info(f"{'Total success ':.<45} {self.success_count}")


def verify_storage(
    sbs_list: list[SpecificBackupSelection],
    local_compare: bool,
    local_compare_root_location: str,
):
    verify_specific_backup_selection_list(sbs_list=sbs_list)
    with Verify(
        verify_selections=sbs_list,
        local_compare=local_compare,
        local_compare_root_location=local_compare_root_location,
    ) as verify:
        verify.verify_files()
    return not verify.is_failed


def handle_verify(args):
    logging.debug(f"handle_verify")
    source_specifiers: list[str] = args.source_storage_specifiers
    sbs_list_list = user_specifiers_to_selections(specifiers=source_specifiers)
    for sbs_list in sbs_list_list:
        sbs_fi_count = 0
        for sbs in sbs_list:
            sbs_fi_count += len(sbs.selected_fi)
        logging.info(
            f"Will verify {sbs_fi_count} files in '{sbs_list[0].storage_def_name}'"
        )

    # No selections is failure.
    is_overall_success = len(sbs_list_list) > 0
    for sbs_list in sbs_list_list:
        is_this_successful = verify_storage(
            sbs_list=sbs_list,
            local_compare=args.compare,
            local_compare_root_location=args.compare_root,
        )
        if not is_this_successful:
            is_overall_success = False
    if is_overall_success:
        logging.info(f"Finished... no errors detected.")
        return 0
    else:
        logging.info(f"Finished... errors were detected.")
        return 1



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Main entry point, argument parsing.
"""
# pylint: disable=missing-class-docstring
# pylint: disable=too-many-lines
import argparse
import logging
import multiprocessing
from typing import Union

from .constants import *
from .exception import (
    AtbuException,
    InvalidCommandLineArgument,
    InvalidStateError,
    QueueListenerNotStarted,
    SingletonAlreadyCreated,
    exc_to_string,
)
from .mp_global import (
    deinitialize_logging,
    get_verbosity_level,
    global_init,
    initialize_logging_basic,
    initialize_logging,
    remove_created_logging_handlers,
    remove_root_stream_handlers,
)
from ..backup.backup_core import (
    BACKUP_COMPRESSION_CHOICES,
    BACKUP_COMPRESSION_DEFAULT,
)
from ..backup.backup_cmdline import handle_backup
from ..backup.restore import handle_restore, handle_decrypt
from ..backup.verify import handle_verify
from ..backup.recover import handle_recover
from ..backup.list_items import handle_list
from ..backup.creds_cmdline import handle_creds
from .file_info import (
    ATBU_PERSISTENT_INFO_EXTENSION,
    CHANGE_DETECTION_CHOICES,
    CHANGE_DETECTION_TYPE_DATESIZE,
    CHANGE_DETECTION_TYPE_DIGEST,
    CHANGE_DETECTION_TYPE_FORCE,
)
from ..persisted_info.update_digests import handle_update_digests
from ..persisted_info.database import handle_savedb
from ..persisted_info.diff import handle_diff, DIFF_COMMAND_CHOICES
from .hasher import DEFAULT_HASH_ALGORITHM, GlobalHasherDefinitions


class BlankLinesHelpFormatter(argparse.HelpFormatter):
    def _split_lines(self, text: str, width: int) -> list[str]:
        return super()._split_lines(text, width) + [""]

    @staticmethod
    def insert_blank_line(parser):
        parser.add_parser("", formatter_class=BlankLinesHelpFormatter, help="")


#
# Common help strings used by argparse configuration further below.
#
storage_def_specifier_help = """
<storage-def-specifier>:
    <storage-def-specifier> is at least a storage specifier, possibly followed by backup and file selectors.
    For the case of a backup, it is simply a storage specifier. For the case of restore, verify and some list
    commands, it is usually a storage specifier, a backup selector, one or more file selectors, all of which
    determine what storage files to restore, verify, list, etc.
        <storage-def-specifier> is shorthand for the following...
        <storage-specifier> [backup:<backup_specifier> [files:<file_specifier> [files:<file_specifier>] ... ] ] ...
    Not all commands allow all portions of a <storage-def-specifier>. See command-specific help for clarification.
    Details:
        <storage_specifier>: storage-def:<storage-definition-name>
            where
                * <storage-definition-name> is either a folder of a local file system backup or
                  a the name you gave to a cloud storage configuration.
                * 'storage-def' can be abbreviated to 'storage' if desired.
            Examples:
                - storage-def:my_cloud_cfg
                - storage:my_cloud_cfg (same as storage-def:my_cloud_cfg)
                - d:\\MyLocalBackup
        backup:<backup_specifier>
            where
                * <backup_specifier> can be 'last' to indicate the most recent backup.
                * <backup_specifier> can be a glob-like selector such as '*20220509-142129*'.
            Examples:
                - backup:last is the most recent backup.
                - backup:MyBackup-20220509-142129 is a specific backup of MyBackup from May 9, 2022 at 14:21:29.
                - backup:MyBackup-2022* are all backups to MyBackup in the year 2022.
        files:<file_specifier>
            where
                * <file_specifier> is essentially a glob-like pattern indicating files to select.
            Examples:
                - files:* will select all files.
                - files:*.txt will select all files ending with ".txt"
"""

extra_help_subjects = {
    "specifiers": storage_def_specifier_help,
}

console_formatter = logging.Formatter("%(asctime)-15s %(threadName)s %(message)s")


def create_argparse():
    #
    # Root parser
    #
    parser = argparse.ArgumentParser(
        description=f"{ATBU_ACRONUM_U} v{ATBU_VERSION_STRING}",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # For certain operations, this causes the secret to be written in hex to the console.
    # This is for testing purposes. You should never use this option when establishing a
    # private key you are actually going to use.
    # parser.add_argument(
    #     "--show-secrets", action="store_true", default=False, help=argparse.SUPPRESS
    # )

    # Uncomment to allow --debug-server (for use with VS Code pydebug)
    parser.add_argument(
        "--debug-server",
        help=argparse.SUPPRESS,  # "Activate the debug server to listen on specified port, wait for a client connect."
        type=int,
        required=False,
    )

    #
    # Common to all parser
    #
    parser_common = argparse.ArgumentParser(add_help=False)
    parser_common.add_argument(
        "--logfile",
        help="The location of log file path. If not specified, do not log to file.",
    )
    parser_common.add_argument(
        "--loglevel", help="level for logging (DEBUG, INFO, WARNING, ERROR, CRITICAL)."
    )
    parser_common.add_argument(
        "--log-console-detail",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="""When logging to console use detailed format.
""",
    )
    parser_common.add_argument(
        "-v",
        "--verbosity",
        action="count",
        help="""increase verbosity with each usage (i.e., -vv is more verbose than -v).
""",
    )

    #############################################################################################
    #                                backup-related argparse setup                              #
    #############################################################################################

    #
    # Common credential key-type argument.
    #
    parser_key_type = argparse.ArgumentParser(add_help=False)
    parser_key_type.add_argument(
        "key_type",
        choices=[CRED_KEY_TYPE_STORAGE, CRED_KEY_TYPE_ENCRYPTION],
        help="What key/password to set (i.e., cloud storage, backup encryption, etc.).",
    )

    #
    # Common credential filename argument.
    #
    parser_credential_filename = argparse.ArgumentParser(add_help=False)
    parser_credential_filename.add_argument(
        "filename",
        help="The path to the credential file.",
    )

    #
    # Common destination file overwrite argument.
    #
    parser_file_ovewrite = argparse.ArgumentParser(add_help=False)
    parser_file_ovewrite.add_argument(
        "--overwrite",
        help="Overwrite the destination file if it already exists.",
        action=argparse.BooleanOptionalAction,
        default=False,
    )

    # Add a subparser to act as a heading with blank lines before/after.
    subparsers = parser.add_subparsers(
        help=f"""""",
    )

    # Add a subparser to act as a heading with blank lines before/after.
    BlankLinesHelpFormatter.insert_blank_line(subparsers)
    subparsers.add_parser(
        "",
        parents=[parser_common],
        help=f"""Backup/Restore/Verify sub-commands
-----------------------------------
""",
    )
    BlankLinesHelpFormatter.insert_blank_line(subparsers)

    #
    # 'backup' subparser.
    #
    parser_backup = subparsers.add_parser(
        "backup",
        formatter_class=argparse.RawTextHelpFormatter,
        help="Backup files to a local file system folder or the cloud.",
        parents=[parser_common],
    )
    group_backup_operation = parser_backup.add_mutually_exclusive_group(required=True)
    group_backup_operation.add_argument(
        f"-{ATBU_BACKUP_TYPE_FULL[0]}",
        f"--{ATBU_BACKUP_TYPE_FULL}",
        help="""Perform a full backup. All files will be backed up regardless of changes.
""",
        action="store_true",
    )
    group_backup_operation.add_argument(
        f"-{ATBU_BACKUP_TYPE_INCREMENTAL[0]}",
        f"--{ATBU_BACKUP_TYPE_INCREMENTAL}",
        help="""Perform an incremental backup. Files whose modified date, size in bytes have changed
since the prior backup.
""",
        action="store_true",
    )
    group_backup_operation.add_argument(
        f"--{ATBU_BACKUP_TYPE_INCREMENTAL_PLUS_SHORT_CMD_OPT}",
        f"--{ATBU_BACKUP_TYPE_INCREMENTAL_PLUS}",
        help="""Perform a comprehensive incremental backup. Files whose modified date, size in bytes,
or hash have changed since the prior backup. This reqiures generating the hash of all
source files which can be intensive and time consuming. Generally, it may be a good
idea, depending on the number of source files, to perform an incremental plus
backup once a while to ensure maximum integrity.
""",
        action="store_true",
    )
    parser_backup.add_argument(
        "source_dirs",
        metavar="source-dirs",
        type=str,
        nargs="+",
        help=f"""One more locations to backup.""",
    )
    parser_backup.add_argument(
        "dest_storage_specifier",
        metavar="dest-storage-specifier",
        type=str,
        help=f"""The destination to backup. The destiation is a <storage-def-specifier>.
To learn more about specifying storage definitions, backups, and files,
see '{ATBU_PROGRAM_NAME} help specifiers' for details.
""",
    )
    parser_backup.add_argument(
        "--dedup",
        choices=[
            ATBU_BACKUP_DEDUPLICATION_TYPE_DIGEST,
            ATBU_BACKUP_DEDUPLICATION_TYPE_DIGEST_EXT,
        ],
        nargs=1,
        help=f"""For --incremental-plus backups only, employ deduplication given the designated
deduplication method:

    '{ATBU_BACKUP_DEDUPLICATION_TYPE_DIGEST}':     This option causes the backup to consider a file unchanged if its
                  digest/date/size match that of an already backed up file, regardless
                  of the other file's path or filename.

    '{ATBU_BACKUP_DEDUPLICATION_TYPE_DIGEST_EXT}': This option is the same as '{ATBU_BACKUP_DEDUPLICATION_TYPE_DIGEST}' except a file must have an extension
                  which matches the already backed up file in order to be considered a
                  duplicate of that file.
                  
                  In all cases, if a deduplication match is detected, the file being
                  considered for backup not backed up but is instead marked a duplicate.
                  During restore, a duplicate resolution process will take place,
                  re-associating a duplicate with a physically backed up file matching
                  the same characteristics used to determine it was a duplicate. Generally
                  speaking, a {DEFAULT_HASH_ALGORITHM} hash match will very rarely have a collision. The
                  additional checks, such as date/size/extension are additional checks
                  which may help mitigate a rare hash collision.

""",
    )
    parser_backup.add_argument(
        "-e",
        "--exclude",
        default=[],
        nargs="+",
        help=f"""Specify one or more glob-like patterns that, if any are matched to a file's path,
the file will not be included in the backup.
""",
    )
    parser_backup.add_argument(
        "--detect-bitrot",
        "--dbr",
        "--detect-sneaky-corruption",
        "--dsc",
        action=argparse.BooleanOptionalAction,
        default=True,
        help=f"""Enabled by default, reports potential bitrot (aka "sneaky corruption") as an error.
Requires use of --incremental-plus (--ip). Use --no-detect-bitrot (--no-dbr) to
squelch reporting as an error (still reported informationally). Bitrot or so-called
sneaky corruption is detected when a file's date/time and size have remained the
same since the last backup, but the digests are different.""",
    )
    parser_backup.add_argument(
        "-z",
        "--compression",
        choices=BACKUP_COMPRESSION_CHOICES,
        help=f"""Set the backup compression level. The default is '{BACKUP_COMPRESSION_DEFAULT}'.
""",
    )
    parser_backup.set_defaults(func=handle_backup)

    #
    # 'restore' subparser.
    #
    parser_restore = subparsers.add_parser(
        "restore",
        formatter_class=argparse.RawTextHelpFormatter,
        parents=[parser_common],
        help="Restore selected files from a backup.",
        description=f"""Restore selected files from a backup.

Examples:

    {ATBU_PROGRAM_NAME} restore d:\\MySourceBackup backup:last files:* c:\\MyRestoreDirectory
        From the last backup to d:\\MySourceBackup, restore all files to c:\\MyRestoreDirectory
        In this case, d:\\MySourceBackup is a local backup (located on d: drive) and not a cloud backup.

    {ATBU_PROGRAM_NAME} restore storage:my_cloud_backup backup:last files:* c:\\MyRestoreDirectory
        From the last backup to storage:my_cloud_backup, restore all files to c:\\MyRestoreDirectory
        In this case, storage:my_cloud_backup is a cloud backup.    
""",
    )
    parser_restore.add_argument(
        "source_storage_specifiers",
        metavar="source-storage-specifiers",
        type=str,
        nargs="+",
        help=f"""The source selections to restore from. To learn more about specifying storage definitions,
backups, and files, see '{ATBU_PROGRAM_NAME} help specifiers' for details.
""",
    )
    parser_restore.add_argument(
        "restore_dir",
        metavar="restore-dir",
        type=str,
        help=f"""The destination directory for the restored files.""",
    )
    parser_restore.add_argument(
        "--overwrite",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="""If specified, allows overwriting already existing destination files.""",
    )
    parser_restore.add_argument(
        "--auto-mapping",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="""Use --no-auto-mapping to disable automatic restore path mapping. By default,
paths mapped to allow restoring with minimal path components relative
to the restore directory. When that is not desirable, use --no-auto-mapping.
""",
    )
    parser_restore.set_defaults(func=handle_restore)

    #
    # 'verify' subparser.
    #
    parser_verify = subparsers.add_parser(
        "verify",
        help="Verify selected files from a backup.",
        description=f"""Verify selected files from a backup.

Examples:

    {ATBU_PROGRAM_NAME} verify d:\\MySourceBackup backup:last files:*
        From the last backup to d:\\MySourceBackup, verify all files.
        In this case, d:\\MySourceBackup is a local backup (located on
        d: drive) and not a cloud backup.

    {ATBU_PROGRAM_NAME} verify storage:my_cloud_backup backup:last files:*
        From the last backup to storage:my_cloud_backup, verify all files
        In this case, storage:my_cloud_backup is a cloud backup.

To learn more about specifying a storage definitions, backups,
and files, see '{ATBU_PROGRAM_NAME} help specifiers'
""",
        formatter_class=argparse.RawTextHelpFormatter,
        parents=[parser_common],
    )
    parser_verify.add_argument(
        "source_storage_specifiers",
        metavar="source-storage-specifiers",
        type=str,
        nargs="+",
        help=f"""The source selections to restore from. To learn more about specifying storage definitions,
backups, and files, see '{ATBU_PROGRAM_NAME} help specifiers' for details.
""",
    )
    parser_verify.add_argument(
        "--compare",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="""
If specified, performs comparison with local copy of the file. If not specified, only the digests
are calculated/compared (default).
""",
    )
    parser_verify.add_argument(
        "--compare-root",
        type=str,
        help="""
When --compare is specified, --compare-root, if specified, is folder under which are the
local files used for comparison. If --compare-root is not specified, the file's original
backup location is used.
""",
    )
    parser_verify.set_defaults(func=handle_verify)

    #
    # 'list' subparser.
    #
    parser_list = subparsers.add_parser(
        "list",
        help="List storage definitions, backups, and files relating to backups.",
        parents=[parser_common],
    )
    parser_list.add_argument(
        "specifiers",
        metavar="[<source> <specific-backup> <file-selections>]",
        type=str,
        nargs="*",
        help=f"""
The source backup specifiers for the list.
Examples:
    {ATBU_PROGRAM_NAME} list
        List all known storage definitions.
    {ATBU_PROGRAM_NAME} list d:\\MySourceBackup backup:last files:*
        From the last backup to d:\\MySourceBackup, list all files.
        In this case, d:\\MySourceBackup is a local backup (located on d: drive) and not a cloud backup.
    {ATBU_PROGRAM_NAME} list storage:my_cloud_backup backup:last files:*
        From the last backup to storage:my_cloud_backup, list all files
        In this case, storage:my_cloud_backup is a cloud backup.
To learn more about specifying a storage definitions, backups, and files, see '{ATBU_PROGRAM_NAME} help specifiers'
""",
    )
    parser_list.set_defaults(func=handle_list)

    #
    # Common storage-def-specifier argument for creds subparsers.
    #
    parser_creds_storage_def_specifier = argparse.ArgumentParser(add_help=False)
    parser_creds_storage_def_specifier.add_argument(
        "storage_def",
        metavar="<storage-def-specifier>",
        type=str,
        help=f"""Backup storage definition to affect. This can be a storage definition name or file
path to file system storage location. See '{ATBU_PROGRAM_NAME} help specifiers'.
""",
    )

    #
    # 'creds' subparser.
    #
    parser_creds = subparsers.add_parser(
        "creds",
        formatter_class=argparse.RawTextHelpFormatter,
        help=f"""Create/delete backup configurations, backup/restore/set credentials.""",
        parents=[parser_common],
    )
    subparser_creds = parser_creds.add_subparsers(
        dest="subcmd",
    )

    #
    # creds create-storage-def:
    #

    # pylint: disable=unused-variable
    cred_create_storage_def_parser = subparser_creds.add_parser(
        "create-storage-def",
        parents=[parser_creds_storage_def_specifier, parser_common],
        formatter_class=argparse.RawTextHelpFormatter,
        help=f"Create a backup storage definition.",
        description=f"""where

    <interface>    <'filesystem','libcloud'|'google'>
    <provider>     <'filesystem'|'azure_blobs'|'google_storage'>
    <container>    The cloud storage container or bucket name.
    <key>          storage key
    <secret>       storage secret
    [,k=v[,k=v ...]] extra parameters as needed: name1=value1,name2=value2,...

Create a storage definition. No spaces allowed as shown. If <secret> evaluates to an
existing filename with a .json extension, the path to that file will be saved as the
crednetial (i.e., an OAuth2 service account file). For some configurations, it may be
easier or necessary to edit the config .json file directly, adding whatever driver
arguments you wish.

Examples:

    {ATBU_PROGRAM_NAME} creds create-storage-def my-cloud-backup libcloud azure_blobs key=<key>,secret=<secret>
    {ATBU_PROGRAM_NAME} creds create-storage-def my-cloud-backup google google_storage key=<client_email>,secret=<path_to_OAuth2.json>

""",
    )
    cred_create_storage_def_parser.add_argument(
        "interface",
        choices=[
            CONFIG_INTERFACE_TYPE_FILESYSTEM,
            CONFIG_INTERFACE_TYPE_LIBCLOUD,
            CONFIG_INTERFACE_TYPE_GOOGLE,
        ],
        help="""The interface (API) the storage definition should use.

""",
    )
    cred_create_storage_def_parser.add_argument(
        "provider",
        metavar="<libcloud_provider_id>",
        help=f"""The storage provider name. This can be '{CONFIG_INTERFACE_TYPE_FILESYSTEM}' for a local
directory, or a provider supported by the chosen interface. For a cloud provider, if
interface is '{CONFIG_INTERFACE_TYPE_LIBCLOUD}' you will typically specify a libcloud provider string such as
'azure_blobs' or 'google_storage'. For provider '{CONFIG_INTERFACE_TYPE_GOOGLE}' use interface 'google_storage'.

""",
    )
    cred_create_storage_def_parser.add_argument(
        "container",
        metavar="<container-name>",
        help=f"""The container name to use. If the name ends with '*', then this parameter acts as a prefix
in the search for an available name which, if found, is then created. A unique
UUID is appended to achieve the goal of finding a name. If a single name without '*'
is specified, it must either already exist or you can use --create-container to
attempt to create that one name.

""",
    )
    cred_create_storage_def_parser.add_argument(
        "driver_params",
        metavar="key=<key>,secret=<secret>[,k=v[,k=v ...]]",
        help=f"""The parameters for the API driver. For example, these are passed to libcloud to
create a driver to access the API. Generally, you must specify at least a key and secret, but
you can specify any libcloud driver arguments.

""",
    )
    cred_create_storage_def_parser.add_argument(
        "--create-container",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="""Attempt to create the container. If you do not specify this, you must
use your cloud provider's user interface to create the container.
""",
    )
    cred_create_storage_def_parser.add_argument(
        "--include-iv",
        "--iiv",
        action=argparse.BooleanOptionalAction,
        default=True,
        help=f"""The default is --include-iv which stores the encryption initialization vector (IV) in the
clear with the encrypted file in the backup storage location. It is recommended you do not
override the default without which you cannot recover encrypted files should you
fail to properly backup your locally stored backup information (which includes the IVs).
""",
    )

    #
    # creds delete-storage-def:
    #
    cred_delete_storage_def_parser = subparser_creds.add_parser(
        "delete-storage-def",
        help="Delete a storage definition.",
        parents=[parser_creds_storage_def_specifier, parser_common],
    )
    cred_delete_storage_def_parser.add_argument(
        "--force",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="Delete without a confirmation prompt.",
    )
    cred_delete_storage_def_parser.add_argument(
        "--delete-backup-info",
        "--dbi",
        action=argparse.BooleanOptionalAction,
        help="""If you specify --no-delete-backup-info (--no-dbi), backup information files
will not be deleted. By default, a storage definition's backup informatiion
files are deleted along with the storage definition itself. If this option
is used without --force, prompting for deleting backup information files
will be skipped (i.e., user only prompted once re storage defintiion).
""",
    )

    #
    # creds export:
    #
    cred_export_cred_parser = subparser_creds.add_parser(
        "export",
        help="Export a storage definition and its secrets to a text file (WARNING: clear text file!).",
        parents=[
            parser_creds_storage_def_specifier,
            parser_credential_filename,
            parser_file_ovewrite,
            parser_common,
        ],
    )

    #
    # creds import:
    #
    cred_import_cred_parser = subparser_creds.add_parser(
        "import",
        help="Import a storage definition and its secrets from a text file.",
        parents=[
            parser_creds_storage_def_specifier,
            parser_credential_filename,
            parser_common,
        ],
    )
    cred_import_cred_parser.add_argument(
        "--create-config",
        action=argparse.BooleanOptionalAction,
        default=False,
        help=f"""Create a {ATBU_DEFAULT_CONFIG_FILE_NAME} configuration if one does not exist.""",
    )
    cred_import_cred_parser.add_argument(
        "--prompt",
        action=argparse.BooleanOptionalAction,
        default=True,
        help=f"""
If --no-prompt is specified, prompts will be skipped. If you specify
to import a backed up configuration, if the same configuration is
already present, it will be overwritten. Ovewriting will wipe out any
existing keys for that storage definition. If keys are different from
the imported configuration, you could lose overwritten keys. It is
recommended you backup configurations before overwriting them by using
'{ATBU_PROGRAM_NAME} creds export ...' first.
""",
    )

    #
    # creds set-password:
    #
    creds_set_password_parser = subparser_creds.add_parser(
        "set-password",
        help="Set character password for specified storage definition.",
        parents=[parser_creds_storage_def_specifier, parser_key_type, parser_common],
    )
    creds_set_password_parser.add_argument(
        "password",
        help="The password to set.",
        nargs="?",
    )

    #
    # creds set-password-filename:
    #
    creds_set_password_filename_parser = subparser_creds.add_parser(
        "set-password-filename",
        help="Set the storage provider password to the specified file name (i.e., OAuth2 .json file).",
        parents=[parser_creds_storage_def_specifier, parser_common],
    )
    creds_set_password_filename_parser.add_argument(
        "filename",
        help="The file name to set.",
    )

    #
    # creds set-password-envvar:
    #
    creds_set_password_envvar_parser = subparser_creds.add_parser(
        "set-password-envvar",
        help="Set password to environment variable name which will point to a file (i.e., credential .json file).",
        parents=[parser_creds_storage_def_specifier, parser_common],
    )
    creds_set_password_envvar_parser.add_argument(
        "env_var",
        help="The environment variable name to set.",
    )
    parser_creds.set_defaults(func=handle_creds)

    #
    # 'restore' subparser.
    #
    parser_recover = subparsers.add_parser(
        "recover",
        parents=[parser_common],
        formatter_class=argparse.RawTextHelpFormatter,
        help="""Recover your backup information files.""",
        description=f"""
This option allows you to recovery backup information from the backup storage, cloud or
local file system. You would do this if your local information has become lost, corrupted
or is otherwise unavailable. You still require your backup's private key, so you should
always make sure you backup (export) your backup's configuration and store it in a safe
place. Without that, you cannot use this recover option.

For more info on exporting your configuration, see '{ATBU_PROGRAM_NAME} creds export ...'
""",
    )
    parser_recover.add_argument(
        "storage_def_cred_cfg",
        metavar="<exported-config-filename> | <storage-def-specifier> [exported-config-filename]",
        nargs="+",
        type=str,
        help=f"""
You can specify the exported config file alone if the recovery
is for a cloud provider storage as the file itself has the
storage definition name, keys etc. needed for recovery.

If you just wish to recover backup information for an already
existing storage def configuration, specify the storage def
name alone, or a path to a filesystem storage definition path.

If you wish to both import credentials and recover backup
information for a filesystem storage, specify the definition's
path and the config filename to import.

The storage definition/credential backup config file is
created by exporting using '{ATBU_PROGRAM_NAME} creds export ...'. For good
measure, you should always backup <home>/{ATBU_DEFAULT_CONFIG_DIR_NAME}/{ATBU_DEFAULT_BACKUP_INFO_SUBDIR}
before performing backup info recovery in case you are uncertain
what you are doing, or otherwise later desire to use or look at
what you had before recovery.
""",
    )
    parser_recover.add_argument(
        "--prompt",
        action=argparse.BooleanOptionalAction,
        default=True,
        help=f"""
If --no-prompt is specified, prompts will be skipped. If you specify
to import a backed up configuration, if the same configuration is
already present, it will be overwritten. Ovewriting will wipe out any
existing keys for that storage definition. If keys are different from
the imported configuration, you could lose overwritten keys. It is
recommended you backup configurations before overwriting them by using
'{ATBU_PROGRAM_NAME} creds export ...' first.
""",
    )
    parser_recover.set_defaults(func=handle_recover)

    #
    # 'decrypt' subparser.
    #
    parser_decrypt = subparsers.add_parser(
        "decrypt",
        formatter_class=argparse.RawTextHelpFormatter,
        parents=[parser_common],
        help="Decrypt storage files directly.",
        description=f"""Decrypt storage files directly. This is generally a command used by someone with technical expertise who may have
a need to decrypt backup storage files without needing to download them. For normal usage/users, this command can
be ignored.

This command might be used in a case where you have the direct raw storage files and you wish to decrypt them. For
example, if you were to download or otherwise get a copy of all your encrypted files stored in the cloud, where
downloading was not required, but you needed to decrypt those files locally in your environment.

Another example might be to sanity check that you can decrypt your storage files without requiring download, perhaps
for prep and confidence that the first example above can be achieved. This might be performed as part of disaster
recovery preparation.

To use this command, you still require your encryption private key so a storage definiton with that key needs to be
configured.

Examples:

    {ATBU_PROGRAM_NAME} decrypt storage:my_cloud_backup c:\\EncryptedFiles c:\\RestoreLocation
        This would use the private encryption key of my_cloud_backup to decrypt files located
        in c:\\EncryptedFiles to c:\\RestoreLocation.
""",
    )
    parser_decrypt.add_argument(
        "private_key_storage_specifier",
        metavar="private-key-storage-specifier",
        type=str,
        help=f"""The storage definition whose private key should be used for decryption.
""",
    )
    parser_decrypt.add_argument(
        "source_files_dir",
        metavar="source-files-dir",
        type=str,
        help=f"""The source directory with optional glob-like pattern specifying the files to decrypt.
If no glob pattern is present, ** is used to select all files. Regardless of the files
found, only those files ending with either {ATBU_FILE_BACKUP_EXTENSION_ENCRYPTED} or {ATBU_FILE_BACKUP_EXTENSION} are considered.""",
    )
    parser_decrypt.add_argument(
        "restore_dir",
        metavar="restore-dir",
        type=str,
        help=f"""The destination directory for the restored files.""",
    )
    parser_decrypt.add_argument(
        "--overwrite",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="""If specified, allows overwriting already existing destination files.""",
    )
    parser_decrypt.set_defaults(func=handle_decrypt)

    #### decrypt END

    # Add a subparser to act as a heading with blank lines before/after.
    BlankLinesHelpFormatter.insert_blank_line(subparsers)
    subparsers.add_parser(
        "",
        parents=[parser_common],
        help=f"""Persistent file information-related sub-commands (unrelated to backup/restore)
------------------------------------------------------------------------------
""",
    )
    BlankLinesHelpFormatter.insert_blank_line(subparsers)

    #############################################################################################
    #                           Persistence-related argparse setup                              #
    #############################################################################################

    #
    # Common to subparsers saving/overwriting .atbu/diff databases.
    #

    #
    # Common to update and diff subparsers.
    #
    common_change_detection_type = argparse.ArgumentParser(add_help=False)
    common_change_detection_type.add_argument(
        "--change-detection-type",
        "--cdt",
        choices=CHANGE_DETECTION_CHOICES,
        default=CHANGE_DETECTION_TYPE_DATESIZE,
        help=f"""For operations where digests can be updated, this option specifies the
method to use when determining if a file has changed, where persistent
info needs updating. The options are:
    '{CHANGE_DETECTION_TYPE_DATESIZE}' for file system date or size changes (default).
    '{CHANGE_DETECTION_TYPE_DIGEST}' for digest check (requires time/cpu for re-gen of file digest).
    '{CHANGE_DETECTION_TYPE_FORCE}' will for an update to all files without the need for checks.
""",
    )
    common_update_stale = argparse.ArgumentParser(add_help=False)
    common_update_stale.add_argument(
        "--update-stale",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="""Specifying --no-update-stale disables the default of updating stale
persistent information. If you specify --no-update-stale, the operation
(such as diff) may be performed using stale persisted information which
may or may not be acceptable depending on your situation. Generally, the
default is what you want because it updates stale information when a
change is detected as determined by --change-detection-type.""",
    )

    #
    # Update digests subparser.
    #
    parser_update_digests = subparsers.add_parser(
        "update-digests",
        parents=[common_change_detection_type, parser_common],
        formatter_class=argparse.RawTextHelpFormatter,
        help="Update persistent file information in the specified directories.",
        description=f"""
Generate or update digests for all files in specified directories, placing the persistent
information in sidecar {ATBU_PERSISTENT_INFO_EXTENSION} files.

WARNING: when using the --persist-type option 'per-file' or 'both', this command creates a new
<filename>{ATBU_PERSISTENT_INFO_EXTENSION} file for all encountered files. By contrast, the 'per-dir' option (the default)
creates a single database at the root of the specified location.

""",
    )
    parser_update_digests.add_argument(
        "locations",
        metavar="[per-file:|pf:|per-dir:|pd:|per-both:|pb:]|<location>",
        nargs="+",
        type=str,
        help=f"""One or more local file system directories whose persistent information to update,
each optionally prefixed with the desired file info persistence to use (default is per-dir).
Note, there is a space after per-file:, per-dir:, and any location. For example,
'per-file: c:\\SomeLocationThatShouldUsePerFile'.
""",
    )
    parser_update_digests.set_defaults(func=handle_update_digests)

    #
    # 'diff' subparser.
    #
    parser_diff = subparsers.add_parser(
        "diff",
        formatter_class=argparse.RawTextHelpFormatter,
        help="Perform a diff of persistent file information of two different directories.",
        description="""Perform a compare of two locations, location A and B, each which has persistent information. Either
location A or B can be a directory or a persistent file information database. The diff produces a
result of A less B, meaning files A that are not in B. Optionally, you can choose to have an action
of removing or moving discovered duplicates.
""",
        parents=[parser_common, common_update_stale, common_change_detection_type],
    )
    parser_diff.add_argument(
        "locations",
        metavar="[per-file:|pf:|per-dir:|pd:|per-both:|pb:]|<location>",
        nargs="+",
        type=str,
        help=f"""You must specify this twice, once each for location A and B, optionally prefixing
either with the desired file info persistence to use (default is per-dir).
Note, there is a space after per-file:, per-dir:, and any location. For example,
'per-file: c:\\SomeLocationThatIsUsingPerFile'.
""",
    )
    parser_diff.add_argument(
        "--relpath-match",
        action="store_true",
        default=False,
        help=f"""If specified, files are only considered to match between location A and
location B if their relative paths match. Relative paths are derived by
removing the top-level location path under which a file was found. By
default, relative paths do not need ot match, only the digest between
the two locations must patch.
""",
    )
    parser_diff.add_argument(
        "--action",
        choices=DIFF_COMMAND_CHOICES,
        help=f"""Perform an action, either remove or move, on files in location A for
which there are duplicates in location B.""",
    )
    parser_diff.add_argument(
        "--move-destination",
        "--md",
        type=str,
        help="Directory for destination when moving duplicates as part of diff.",
    )
    parser_diff.add_argument(
        "--whatif",
        action="store_true",
        default=False,
        help="Show what files would be removed but do not actually remove them.",
    )
    parser_diff.set_defaults(func=handle_diff)

    #
    # 'savedb' subparser.
    #
    parser_savedb = subparsers.add_parser(
        "save-db",
        formatter_class=argparse.RawTextHelpFormatter,
        help="Save persistent file information from or more directories into a single json db.",
        description=f"""Save a json database from all persistent information created by the
'{ATBU_PROGRAM_NAME} update-digests' command.

""",
        parents=[parser_common, common_update_stale, common_change_detection_type],
    )
    parser_savedb.add_argument(
        "locations",
        metavar="[--per-file|--pf|--per-dir|--pd|--per-both|--pb] <location1>",
        nargs="+",
        help=f"""The locations whose persistent file information should be placed
into a newly saved database file.
""",
    )
    parser_savedb.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="When requesting to write a hash db file, overwrite any existing db file.",
    )
    parser_savedb.add_argument(
        "--database",
        "--db",
        type=str,
        required=True,
        help="Path of the persistent file information json database.",
    )
    parser_savedb.set_defaults(func=handle_savedb)

    # Add a subparser to act as a heading with blank lines before/after.
    BlankLinesHelpFormatter.insert_blank_line(subparsers)
    subparsers.add_parser(
        "",
        parents=[parser_common],
        help=f"""Other sub-commands
------------------
""",
    )
    BlankLinesHelpFormatter.insert_blank_line(subparsers)

    #############################################################################################
    #                specific help argparse setup (i.e., for atbu help <subject>                #
    #############################################################################################

    parser_help = subparsers.add_parser(
        "help",
        formatter_class=argparse.RawTextHelpFormatter,
        help=f"""Show help on specific subjects:
    '{ATBU_PROGRAM_NAME} help -h' to see available subjects.""",
    )
    parser_help.add_argument(
        "help_subject",
        choices=[k for k in extra_help_subjects],
        nargs=1,
        help=f"""The subject to get detailed help on:
    '{ATBU_PROGRAM_NAME} help specifiers' - get detailed help on storage definition specifiers used by backup/restore commands.
""",
    )

    return parser


def wait_for_debugger_attach(debug_server_port: Union[str, int, list]):
    if debug_server_port is None:
        raise InvalidCommandLineArgument(f"The debugger port must be specified.")
    if isinstance(debug_server_port, list):
        debug_server_port = debug_server_port[0]
    if isinstance(debug_server_port, str):
        try:
            debug_server_port = int(debug_server_port)
        except Exception as ex:
            raise InvalidCommandLineArgument(
                f"Parsing error: "
                f"The value '{debug_server_port}' is an invalid port number."
            ) from ex
    if debug_server_port < 0 or debug_server_port > 65535:
        raise InvalidCommandLineArgument(
            f" The value '{debug_server_port}' is an invalid port number."
        )
    try:
        import debugpy  # pylint: disable=import-outside-toplevel

        debugpy.listen(debug_server_port)
        print(f"Waiting for the debugger to attach via port {debug_server_port}...")
        debugpy.wait_for_client()
        debugpy.breakpoint()
        print(f"Debugger connected.")
    except ModuleNotFoundError as ex:
        raise InvalidStateError(
            f"Cannot 'import debugpy'. Either ensure vscode debugpy is available."
        ) from ex
    except Exception as ex:
        raise InvalidStateError(
            f"Unexpected error. Cannot wait for debugger attach. {exc_to_string(ex)}"
        ) from ex


def main(argv=None):
    multiprocessing.set_start_method("spawn")
    global_init()
    initialize_logging_basic()
    # pdb.set_trace()
    logging.info(f"{ATBU_ACRONYM} - v{ATBU_VERSION_STRING}")

    parser = create_argparse()
    args = parser.parse_args(argv)

    if hasattr(args, "debug_server") and args.debug_server is not None:
        wait_for_debugger_attach(args.debug_server)

    if hasattr(args, "help_subject"):
        for requested_help_subject in args.help_subject:
            if requested_help_subject in extra_help_subjects:
                print(extra_help_subjects[requested_help_subject])
                parser.exit(1)
        raise InvalidStateError(
            f"Expected help subject to match those available: {args.help_subject}"
        )

    verbosity_level = 0
    if hasattr(args, "verbosity") and args.verbosity is not None:
        verbosity_level = args.verbosity

    # Instantiate global singleton GlobalHasherDefinitions.
    try:
        GlobalHasherDefinitions([DEFAULT_HASH_ALGORITHM])
    except SingletonAlreadyCreated:
        logging.warning(f"GlobalHasherDefinitions already created.")

    logfile = None
    loglevel = None
    if hasattr(args, "logfile"):
        logfile = args.logfile
    if hasattr(args, "loglevel"):
        loglevel = args.loglevel
    # If no log file specified and user is creating/updating database, automatically place log file side-by-side with db.
    if (
        not logfile
        and hasattr(args, "func")
        and args.func == handle_savedb  # pylint: disable=comparison-with-callable.
        and hasattr(args, "database")  # pylint: disable=W0143
        and args.database is not None
    ):
        logfile = args.database + ".log"

    log_console_detail = False
    if hasattr(args, "log_console_detail") and args.log_console_detail:
        log_console_detail = args.log_console_detail

    exit_code = 1
    try:
        if hasattr(args, "func"):
            remove_created_logging_handlers()
            remove_root_stream_handlers()
            initialize_logging(
                logfile=logfile,
                loglevel=loglevel,
                verbosity_level=verbosity_level,
                log_console_detail=log_console_detail,
            )
            exit_code = args.func(args)
            if exit_code is None:
                exit_code = 0
        else:
            print(f"I have nothing to do. Try {ATBU_PROGRAM_NAME} -h for help.")
    except AtbuException as err:
        logging.error(f"Failed: {err.message}")
        if get_verbosity_level() > 0:
            logging.error(exc_to_string(err))
            raise
    finally:
        try:
            deinitialize_logging()
        except QueueListenerNotStarted:
            pass
        except Exception as ex:
            print(f"Failure during stop_global_queue_listener. {exc_to_string(ex)}")
    logging.debug(f"{ATBU_PROGRAM_NAME} exit_code={exit_code}")
    return exit_code


if __name__ == "__main__":
    main()



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Classes/functions related to backup-related selections.
"""
from dataclasses import dataclass, field
from fnmatch import fnmatchcase, fnmatch
from glob import iglob
import logging
import os
from pathlib import Path
import platform
import re

from ..common.exception import (
    BackupSelectionError,
    StorageDefinitionNotSpecifiedError,
    StorageDefinitionNotFoundError,
    BackingFileInformationNotFound,
    InvalidStateError,
)
from ..common.constants import *
from .config import (
    AtbuConfig,
    is_existing_filesystem_storage_path,
)
from .backup_core import (
    StorageDefinition,
    SpecificBackupInformation,
    BackupInformationDatabase,
)
from .backup_core import BackupFileInformation

DEFAULT_PLATFORM_EXCLUDE = {
    "Windows": {
        re.compile(os.path.normcase(r"^[a-zA-Z]:\\System Volume Information" "\\\\")),
        re.compile(os.path.normcase(r"^[a-zA-Z]:\\$Recycle.Bin" "\\\\")),
    }
}

SELECTION_SCOPE_BACKUP = "backup"
SELECTION_SCOPE_FILE = "files"
SELECTION_SCOPE_ALL = "all"
SELECTION_SCOPES = [SELECTION_SCOPE_ALL, SELECTION_SCOPE_BACKUP, SELECTION_SCOPE_FILE]

SELECTION_BACKUP_LAST = "last"


class SelectionPattern:
    def __init__(self, raw_command_line_specifier: str):
        parts = raw_command_line_specifier.split(":", maxsplit=1)
        scope = None
        pattern = parts
        if len(parts) > 1:
            scope = parts[0]
            pattern = parts[1]
        if scope not in SELECTION_SCOPES:
            raise BackupSelectionError(
                f"The scope '{scope}' is invalid, must be one of {SELECTION_SCOPES}."
            )
        self.scope = scope
        self.original_pattern = pattern
        self.normcase_pattern = os.path.normcase(pattern)

    @staticmethod
    def is_valid_selection_pattern(arg_to_test: str):
        parts = arg_to_test.split(":", maxsplit=1)
        return len(parts) == 2 and parts[0] in SELECTION_SCOPES


@dataclass
class StorageSelectionInfo:
    location: str
    location_derivation_source_arg: str
    storage_def_name: str
    storage_def: dict
    storage_atbu_cfg: AtbuConfig
    backup_info_dir: Path
    selection_patterns: list[SelectionPattern]


@dataclass
class SpecificBackupSelection:
    storage_def_name: str
    specific_backup_name: str
    specific_backup_info: SpecificBackupInformation
    storage_def: StorageDefinition
    backup_info_dir: Path
    backup_history: BackupInformationDatabase
    sel_info: StorageSelectionInfo
    selected_fi: dict[str, BackupFileInformation] = field(default_factory=dict)


def parse_storage_def_specifiers_patterns(
    raw_arguments: list[str], resolve_storage_def_secrets: bool = False
) -> list[StorageSelectionInfo]:
    sel_info_dict: dict[str, StorageSelectionInfo] = {}
    result: list[StorageSelectionInfo] = []
    default_cfg = AtbuConfig.access_default_config()
    cur_storage_selections: list[StorageSelectionInfo] = None
    for raw_arg in raw_arguments:
        if SelectionPattern.is_valid_selection_pattern(arg_to_test=raw_arg):
            #
            # Valid selection pattern, apply it to cur_storage_selections.
            #
            if not cur_storage_selections:
                raise StorageDefinitionNotSpecifiedError(
                    f"No storage selection to apply pattern '{raw_arg}' to."
                )
            for (
                cur_storage_selection
            ) in cur_storage_selections:  # pylint: disable=not-an-iterable
                cur_storage_selection.selection_patterns.append(
                    SelectionPattern(raw_command_line_specifier=raw_arg)
                )
            continue

        storage_pattern = raw_arg.lower()
        raw_arg_split = raw_arg.split(":", maxsplit=1)
        if (
            raw_arg_split[0]
            == CONFIG_SECTION_STORAGE_DEFINITION_SPECIFIER_PREFIX[
                : len(raw_arg_split[0])
            ]
        ):
            storage_pattern = raw_arg_split[1]

        storage_specifiers = [
            f"{CONFIG_SECTION_STORAGE_DEFINITION_SPECIFIER_PREFIX}:{n}"
            for n in default_cfg.get_storage_def_names(fnmatch_pattern=storage_pattern)
        ]

        if not storage_specifiers:
            if not is_existing_filesystem_storage_path(storage_location=raw_arg):
                raise StorageDefinitionNotFoundError(
                    f"The specifier '{raw_arg}' did not yield any storage definitions or filesystem backups."
                )
            # The raw_arg is a directory, set it to the specifier to allow following
            # to try to find a configuration.
            storage_specifiers = [raw_arg]

        #
        # Fall through, raw_arg is not a selection pattern, should be storage definition specifier...
        #

        # Per-iteration storage selections.
        cur_storage_selections: list[StorageSelectionInfo] = []

        # For each specifier, find a storage, add to cur_storage_selections
        # for extracting any patterns.
        for storage_specifier in storage_specifiers:
            atbu_cfg_to_use: AtbuConfig
            (
                atbu_cfg_to_use,
                storage_def_name,
                storage_def_dict,
            ) = default_cfg.resolve_storage_location(
                storage_location=storage_specifier,
                resolve_storage_def_secrets=resolve_storage_def_secrets,
                create_if_not_exist=False,
            )
            if not atbu_cfg_to_use or not storage_def_name or not storage_def_dict:
                raise StorageDefinitionNotFoundError(
                    f"WARNING: Storage definition for '{storage_specifier}' not found, skipping."
                )

            storage_sel_info: StorageSelectionInfo = sel_info_dict.get(storage_def_name)
            if not storage_sel_info:
                storage_def = StorageDefinition.storage_def_from_dict(
                    storage_def_name=storage_def_name, storage_def_dict=storage_def_dict
                )

                storage_sel_info = StorageSelectionInfo(
                    location=storage_specifier,
                    location_derivation_source_arg=raw_arg,
                    storage_def_name=storage_def_name,
                    storage_def=storage_def,
                    storage_atbu_cfg=default_cfg,
                    backup_info_dir=atbu_cfg_to_use.get_backup_info_dir(),
                    selection_patterns=[],
                )
                # Add to ordered list of StorageSelectionInfo instance results.
                result.append(storage_sel_info)

            # Add to the StorageSelectionInfo instances related to
            # the upcoming SelectionPattern instances (further above).
            # A single StorageSelectionInfo instance could be added to
            # cur_storage_selections twice if perhaps the same storage def
            # was specified twice in the overall options.
            cur_storage_selections.append(storage_sel_info)

    # Resulting list should only have any given storage defintion once,
    # with any patterns extracted per above.
    return result


def get_selections_from_file_info_dict(
    sel_pat: SelectionPattern, nc_path_to_fi_dict: dict[str, BackupFileInformation]
) -> list[BackupFileInformation]:
    r: list[BackupFileInformation] = []
    # Check each normcase'ed path against the glob-like pattern.
    for normcase_path, fi in nc_path_to_fi_dict.items():
        # path and pattern are already normcase so using fnmatchcase is fine.
        if not fnmatchcase(normcase_path, sel_pat.normcase_pattern):
            continue
        # Sanity check state, ensure file is resolved with backing_fi.
        if fi.is_unchanged_since_last and not fi.backing_fi:
            # If unchanged file has unresolved backing_fi, fail.
            raise BackingFileInformationNotFound(
                f"An unchanged since last file has no backing file information: "
                f"fi={fi.path}"
            )
        r.append(fi)
    return r


def get_specific_backup_selections(
    sel_info_list: list[StorageSelectionInfo],
) -> list[list[SpecificBackupSelection]]:
    result: list[list[SpecificBackupSelection]] = []
    history_dict: dict[str, BackupInformationDatabase] = {}
    # For each storage definition being filtered...
    for si in sel_info_list:
        if not si.selection_patterns:
            # No filters, skip.
            # TODO: Should this default to none or * all ? (TBD)
            # For now, create a dummy SpecificBackupSelection to allow
            # querying of storage defintions via this code path.
            specific_backup_sel = SpecificBackupSelection(
                storage_def_name=si.storage_def_name,
                specific_backup_name=None,
                specific_backup_info=None,
                storage_def=si.storage_def,
                backup_info_dir=si.backup_info_dir,
                backup_history=None,
                sel_info=si,
            )
            result.append([specific_backup_sel])
            continue
        # For the si storage definition, extract all selected specific backup instances.
        specific_backup_selections: list[SpecificBackupSelection] = []
        for sp in si.selection_patterns:
            if sp.scope not in [SELECTION_SCOPE_BACKUP]:
                continue
            #
            # Share the same storage definition history with SpecificBackupSelection
            # instances of the same storage definition as follows.
            #
            history_key = f"{si.storage_def_name}:{si.backup_info_dir}"
            backup_history = history_dict.get(history_key)
            if not backup_history:
                backup_history = BackupInformationDatabase.create_from_file(
                    backup_base_name=si.storage_def_name,
                    backup_info_dir=si.backup_info_dir,
                )
            # For each specific backup, from newest to oldest...
            for sbi in backup_history.get_specific_backups():
                # If user specified SELECTION_BACKUP_LAST, always match and
                # select the first most recent specific backup.
                # If the user specified something actual fnmatch criteria and
                # it matches, select the backup.
                # All other cases skip this specific backup.
                if (
                    not sp.original_pattern.lower() == SELECTION_BACKUP_LAST
                    and not fnmatch(sbi.specific_backup_name, sp.normcase_pattern)
                ):
                    continue
                # Match, backup selections.
                specific_backup_sel = SpecificBackupSelection(
                    storage_def_name=si.storage_def_name,
                    specific_backup_name=sbi.specific_backup_name,
                    specific_backup_info=sbi,
                    storage_def=si.storage_def,
                    backup_info_dir=si.backup_info_dir,
                    backup_history=backup_history,
                    sel_info=si,
                )
                specific_backup_selections.append(specific_backup_sel)
                if sp.original_pattern.lower() == SELECTION_BACKUP_LAST:
                    # With selector SELECTION_BACKUP_LAST skip backups
                    # older than the last one (the most recent one).
                    break
        for sp in si.selection_patterns:
            if sp.scope != SELECTION_SCOPE_FILE:
                continue
            specific_backup_sel: SpecificBackupSelection
            # For each selected specific backup, from newest to oldest...
            for specific_backup_sel in specific_backup_selections:
                # Place all specific backup file info into a dict keyed w/normcase path.
                sb_nc_path_to_fi_dict: dict[str, BackupFileInformation] = {}
                for sb_fi in specific_backup_sel.specific_backup_info.all_file_info:
                    sb_nc_path_to_fi_dict[sb_fi.nc_path_without_root] = sb_fi
                # Determine which specific backup file info matches the file selection pattern.
                list_sel_fi = get_selections_from_file_info_dict(
                    sel_pat=sp, nc_path_to_fi_dict=sb_nc_path_to_fi_dict
                )
                # For each matching specific backup file info, if it has not already
                # matched the same path with a newer file info, add it to the selection.
                # There should not be duplicate paths per specific backup.
                # Callers will use functions like get_all_specific_backup_file_info to
                # reduce all selections across all specific backups to a single of duplicates
                # where applicable. That is not what the following does. The following is a
                # sanity check.
                for sel_fi in list_sel_fi:
                    if not specific_backup_sel.selected_fi.get(
                        sel_fi.nc_path_without_root
                    ):
                        specific_backup_sel.selected_fi[
                            sel_fi.nc_path_without_root
                        ] = sel_fi
        if specific_backup_selections:
            result.append(specific_backup_selections)
    return result


def verify_specific_backup_selection_list(sbs_list: list[SpecificBackupSelection]):
    storage_def_name = None
    backup_info_dir = None
    for sbs in sbs_list:
        if not storage_def_name:
            storage_def_name = sbs.storage_def_name
            backup_info_dir = sbs.backup_info_dir
        elif (
            storage_def_name != sbs.storage_def_name
            or backup_info_dir != sbs.backup_info_dir
        ):
            raise InvalidStateError(
                f"All SpecificBackupSelection instances in the list must target the same storage definition."
            )


def get_all_specific_backup_file_info(
    sbs_list: list[SpecificBackupSelection],
) -> list[BackupFileInformation]:
    sb_nc_path_to_fi_dict: dict[str, BackupFileInformation] = {}
    # The sbs_list contains SpecificBackupSelection instances from
    # the most recent to oldest specific backup examined as part of
    # the selection process. For each SpecificBackupSelection instance
    # in that order...
    for sbs in sbs_list:
        # Go through the SpecificBackupSelection instance's items, if an item
        # is not in the result sb_nc_path_to_fi_dict, add it. This means
        # newest BackupFileInformation wins.
        for _, sbs_fi in sbs.selected_fi.items():
            if not sb_nc_path_to_fi_dict.get(sbs_fi.nc_path_without_root):
                sb_nc_path_to_fi_dict[sbs_fi.nc_path_without_root] = sbs_fi
    # Return a list of BackupFileInformation instances sorted by normcase path.
    # The sorting is not essential but during operations it can be nice to see
    # files processed with in an order making sense. Being nonessential, this
    # can be changed if needed.
    return sorted(
        [v for _, v in sb_nc_path_to_fi_dict.items()],
        key=lambda sb_fi: sb_fi.nc_path_without_root,
    )


def user_specifiers_to_selections(
    specifiers: list[str], no_selections_ok: bool = False
) -> list[list[SpecificBackupSelection]]:

    # Get selection information from all specifiers.
    sel_info_list = parse_storage_def_specifiers_patterns(
        raw_arguments=specifiers, resolve_storage_def_secrets=True
    )

    # Validate selection information.
    if not no_selections_ok:
        for sel_info in sel_info_list:
            if len(sel_info.selection_patterns) == 0:
                raise BackupSelectionError(
                    f"You must have at least one storage, backup, and file specifier. "
                    f"Example: storage:<storage_def_name> backup:last files:*"
                )

    # Get selections using selection information.
    sbs_list_list = get_specific_backup_selections(sel_info_list)

    if not no_selections_ok:
        # Validate selections.
        for sbs_list in sbs_list_list:
            file_info_count = 0
            for sbs in sbs_list:
                file_info_count += len(sbs.selected_fi)
            if file_info_count == 0:
                raise BackupSelectionError(
                    f"No selected files for {sbs_list[0].storage_def_name}. "
                    f"You must have at least one storage, backup, and file specifier. "
                    f"Example: storage:<storage_def_name> backup:last files:*"
                )

    # Everything seems OK, return selections to caller.
    return sbs_list_list


def get_storage_defintions_from_sbs_list_list(
    sbs_list_list: list[list[SpecificBackupSelection]],
) -> list[StorageDefinition]:
    storage_def_dict: dict[str, StorageDefinition] = {}
    for sbs_list in sbs_list_list:
        for sbs in sbs_list:
            storage_def_dict[sbs.storage_def_name] = sbs.storage_def
    storage_def_list: list[StorageDefinition] = sorted(
        storage_def_dict.values(), key=lambda sd: sd.storage_def_name
    )
    return storage_def_list


def is_system_excluded_path(path):
    platform_excludes = DEFAULT_PLATFORM_EXCLUDE.get(platform.system())
    if platform_excludes is not None:
        for pe in platform_excludes:
            if pe.match(os.path.normcase(path)):
                return True
    return False


RE_FIND_GLOB_CHARS = re.compile(r"[?\[\]*]")
PATH_SEP_CHARS = r"\/"
RE_CONTAINS_GLOB_CHARS = re.compile(".*[*?[].*")


def get_largest_path_without_wildcards(path: str):
    m = RE_FIND_GLOB_CHARS.search(path)
    if m:
        index = m.start()
        while path[index] not in PATH_SEP_CHARS:
            index -= 1
        if index >= 0:
            path = path[: index + 1]
    return path


def get_dirs_without_wildcards(locations: list[str]):
    results: list[str] = []
    for location in locations:
        location = get_largest_path_without_wildcards(location)
        results.append(os.path.abspath(location))
    return results


def ensure_glob_pattern_for_dir(dir_wc: str):
    """For a directory, if no glob pattern is present, ensure
    the glob "**" pattern is joined.
    """
    nc_dir_wc = os.path.normcase(dir_wc)
    nc_drive = os.path.splitdrive(nc_dir_wc)[0]
    if not RE_CONTAINS_GLOB_CHARS.match(dir_wc) and (
        os.path.isdir(dir_wc) or nc_dir_wc == nc_drive
    ):
        dir_wc = os.path.join(dir_wc, "**")
    return dir_wc


@dataclass
class SourceSpecifierDiscoveries:
    source_dir: str
    source_dir_wc: str
    discovered_paths: list[BackupFileInformation]


def get_local_file_information(
    src_dirs_wc: list[str], exclude_patterns: list[str]
) -> list[BackupFileInformation]:
    r"""Given locations containing one or more paths to search, return a list
    of discovered files.

    Each path in src_dirs_wc may or may not contain glob patterns. If a path in
    src_dirs_wc is a directory without a glob pattern, the '**' glob pattern
    will be appened. For example, "c:\abc" will become "c:\\abc\\**" to
    cause a recursive search of all files.

    This function will return a list of file information each of which will
    have its file_info.discovery_path set to the src_dirs_wc leading to its
    discovery.
    """

    exclude_patterns = list(map(os.path.normcase, exclude_patterns))

    def is_ignored(path):
        if is_system_excluded_path(path):
            logging.debug(f"Ignoring platform-excluded path: {path}")
            return True
        nonlocal exclude_patterns
        if len(exclude_patterns) == 0:
            return False
        path_nc = os.path.normcase(path)
        if any(map(lambda pat: fnmatchcase(name=path_nc, pat=pat), exclude_patterns)):
            logging.info(f"Ignoring user-excluded path: {path}")
            return True
        return False

    # Both src_dirs_wc and src_dirs_no_wc can be correlated by index.
    # Start by sorting caller's selections by length/normcase. Length
    # means discovery by shortest selector wins.
    src_dirs_wc.sort(key=lambda p: (len(p), os.path.normcase(p)))
    src_dirs_no_wc: list[str] = get_dirs_without_wildcards(locations=src_dirs_wc)

    # Ensure no duplicate paths in final selection.
    all_paths = set()

    # Track which src_dirs_wc selector discovered which file path.
    srcspec_to_disc: dict[str, SourceSpecifierDiscoveries] = {}

    # For each caller selection...
    for idx, src_dir_wc in enumerate(src_dirs_wc):

        # Get corresponding path without any glob patterns ("wildcards" or "wc").
        src_dir_no_wc = src_dirs_no_wc[idx]
        src_dir_no_wc_nc = os.path.normcase(src_dir_no_wc)

        # If location is either a directory or drive letter alone without
        # any glob patterns, then add glob pattern '**'.
        nc_location = os.path.normcase(src_dir_wc)
        nc_drive = os.path.splitdrive(nc_location)[0]
        if not RE_CONTAINS_GLOB_CHARS.match(src_dir_wc) and (
            os.path.isdir(src_dir_wc) or nc_location == nc_drive
        ):
            src_dir_wc = os.path.join(src_dir_wc, "**")

        # Search, removing anything already found.
        discovered = set(iglob(pathname=src_dir_wc, recursive=True)).difference(
            all_paths
        )

        # Refine further, retaining only files which are not ignored.
        discovered = [v for v in discovered if os.path.isfile(v) and not is_ignored(v)]

        if len(discovered) == 0:
            # Nothing left, next item.
            continue

        # For the given search root folder (without glob pattern),
        # find an existing SourceSpecifierDiscoveries instance, if
        # non found, create one.
        src_spec_disc = srcspec_to_disc.get(src_dir_no_wc_nc)
        if not src_spec_disc:
            src_spec_disc = SourceSpecifierDiscoveries(
                source_dir=src_dir_no_wc, source_dir_wc=src_dir_wc, discovered_paths=[]
            )
            srcspec_to_disc[src_dir_no_wc_nc] = src_spec_disc

        # Add discovered files to instance, assocating them with
        # root path used to find the selections themselves.
        src_spec_disc.discovered_paths.extend(discovered)

        # Track files added, do not add more than once.
        all_paths.update(discovered)

    # Create results for caller by processing each SourceSpecifierDiscoveries instance.
    disc_list = sorted(
        srcspec_to_disc.values(), key=lambda d: os.path.normcase(d.source_dir)
    )
    file_info_list: list[BackupFileInformation] = []
    # For each caller-specified search path...
    for idx, disc in enumerate(disc_list):
        # For each file discovered by caller's search path...
        for p in disc.discovered_paths:
            if not os.path.normcase(p).startswith(os.path.normcase(disc.source_dir)):
                raise InvalidStateError(
                    f"Expected file to have path starting with discovery source: "
                    f"source_dir_wc={disc.source_dir_wc} "
                    f"file_info.path={p.path}"
                )
            # Create the result list with file info
            # indicating discovery source dir.
            file_info_list.append(
                BackupFileInformation(path=p, discovery_path=disc.source_dir)
            )
    return file_info_list



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Credential and storage-definition-related creation/management.
"""

import json
import os
import logging
import re
from typing import Union
from send2trash import send2trash

from ..common.util_helpers import get_trash_bin_name, prompt_YN
from ..common.mp_global import switch_to_non_queued_logging
from ..common.constants import *
from .config import (
    AtbuConfig,
    is_existing_filesystem_storage_path,
    is_storage_def_name_ok,
    parse_storage_def_specifier,
)
from .credentials import (
    Credential,
    CredentialByteArray,
    prompt_for_password,
    set_password_to_keyring,
)
from ..common.exception import (
    AtbuException,
    ContainerAlreadyExistsError,
    ContainerAutoCreationFailed,
    CredentialRequestInvalidError,
    CredentialTypeNotFoundError,
    InvalidCommandLineArgument,
    InvalidContainerNameError,
    InvalidStateError,
    InvalidStorageDefinitionName,
    PasswordAuthenticationFailure,
    CredentialInvalid,
    StorageDefinitionNotFoundError,
    UnexpectedOperation,
    EncryptionAlreadyEnabledError,
    CredentialSecretDerivationError,
    StorageDefinitionAlreadyExists,
    exc_to_string,
)
from .storage_interface.base import (
    AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR,
    StorageInterfaceFactory,
)


def handle_credential_change(
    operation: CRED_OPERATION_HINT,
    key_to_affect: CRED_KEY_TYPE_HINT,
    atbu_cfg: AtbuConfig,
    storage_def_name: str,
    credential: Union[str, list, Credential],
    debug_mode: bool = False,
):
    if not storage_def_name:
        raise ValueError(f"Invalid backup storage definition: {storage_def_name}")

    #
    # Given key_to_affect, the following dict provides info about what
    # keyring username to use as well as what ATBU config nodes to affect.
    #
    AFFECTED_KEY_TO_INFO = {
        CRED_KEY_TYPE_ENCRYPTION: (
            CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION,
            "encryption/key",
        ),
        CRED_KEY_TYPE_STORAGE: (
            CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD,
            "driver/secret",
        ),
    }

    #
    # Given operation, look up the password type to store.
    #
    OPERATION_TO_PASSWORD_TYPE = {
        CRED_OPERATION_SET_PASSWORD: CONFIG_PASSWORD_TYPE_ACTUAL,
        CRED_OPERATION_SET_PASSWORD_TO_FILENAME: CONFIG_PASSWORD_TYPE_FILENAME,
        CRED_OPERATION_SET_PASSWORD_TO_ENVVAR: CONFIG_PASSWORD_TYPE_ENVVAR,
        CRED_OPERATION_SET_PASSWORD_TO_PRIVATE_KEY: CONFIG_PASSWORD_TYPE_ACTUAL,
    }

    #
    # Determine username and config section given key_to_affect.
    #
    username = AFFECTED_KEY_TO_INFO[key_to_affect][0]
    section_path = AFFECTED_KEY_TO_INFO[key_to_affect][1]
    password_type = OPERATION_TO_PASSWORD_TYPE[operation]

    logging.info(f"Keyring information:")
    logging.info(f"Key={key_to_affect}")
    logging.info(f"Service={storage_def_name}")
    logging.info(f"Username={username}")
    logging.debug(f"password_type={password_type}")
    logging.debug(f"section_path={section_path}")

    if key_to_affect not in [CRED_KEY_TYPE_STORAGE, CRED_KEY_TYPE_ENCRYPTION]:
        raise UnexpectedOperation(f"The the '{key_to_affect}' key is unexpected.")

    #
    # CRED_KEY_TYPE_ENCRYPTION
    #
    if key_to_affect == CRED_KEY_TYPE_ENCRYPTION:

        if operation not in [
            CRED_OPERATION_SET_PASSWORD,
            CRED_OPERATION_SET_PASSWORD_TO_PRIVATE_KEY,
        ]:
            raise UnexpectedOperation(
                f"The operation '{operation}' for the '{key_to_affect}' key is unexpected."
            )

        if not isinstance(credential, Credential):
            raise CredentialTypeNotFoundError(
                f"A Credential type is expected when setting the '{key_to_affect}' secret"
            )

        if operation == CRED_OPERATION_SET_PASSWORD_TO_PRIVATE_KEY:
            # Direct private key, ensure key is present.
            if not credential.is_private_key_ready:
                raise CredentialInvalid(
                    f"Cannot set the '{key_to_affect}' key. "
                    f"The Credential is invalid because the private key is not immediately available"
                )
            # Direct key storage.
            credential_bytes = credential.get_as_bytes(include_key=True)
            if debug_mode:
                print(f"handle_credential_change: key={credential.the_key.hex(' ')}")
                print(f"handle_credential_change: b={credential_bytes}")
        elif operation == CRED_OPERATION_SET_PASSWORD:
            # Private key will be password-protected.
            if not credential.is_private_key_possible:
                raise CredentialInvalid(
                    f"Cannot set the '{key_to_affect}' key. "
                    f"The Credential cannot be used to derive the private key and is therefore invalid."
                )
            credential_bytes = credential.get_as_bytes(
                include_work_factor=True,
                include_salt=True,
                include_password_auth_hash=True,
                include_IV=True,
                include_encrypted_key=True,
            )
            if debug_mode:
                if credential.the_key:
                    print(
                        f"handle_credential_change: key={credential.the_key.hex(' ')}"
                    )
                else:
                    print(f"handle_credential_change: key=<none>")
                print(f"handle_credential_change: b={credential_bytes}")
    #
    # CRED_KEY_TYPE_STORAGE
    #
    elif key_to_affect == CRED_KEY_TYPE_STORAGE:

        if operation not in [
            CRED_OPERATION_SET_PASSWORD,
            CRED_OPERATION_SET_PASSWORD_TO_FILENAME,
            CRED_OPERATION_SET_PASSWORD_TO_ENVVAR,
        ]:
            raise UnexpectedOperation(
                f"The operation '{operation}' for the '{key_to_affect}' key is unexpected."
            )

        if not isinstance(credential, (CredentialByteArray, bytearray, bytes, str)):
            raise CredentialInvalid(
                f"The credential type '{type(credential)}' is not expected when setting the '{key_to_affect}' key."
            )

        credential_bytes = credential
        if isinstance(credential, str):
            credential_bytes = credential.encode("utf-8")
    #
    # Unexpected
    #
    else:
        raise UnexpectedOperation(f"The '{key_to_affect}' is unexpected.")

    set_password_to_keyring(
        service_name=storage_def_name,
        username=username,
        password_type=password_type,
        password_bytes=credential_bytes,
    )

    #
    # Update the configuration to reference the keyring data.
    #
    atbu_cfg.set_config_keyring_mapping(
        storage_def_name=storage_def_name, username=username, section_path=section_path
    )
    atbu_cfg.save_config_file()


def set_encryption_password_wizard(
    atbu_cfg: AtbuConfig,
    storage_def_name: str,
    encryption_already_enabled_ok: bool = False,
    debug_mode: bool = False,
):
    credential: Credential = None
    if atbu_cfg.is_storage_def_configured_for_encryption(
        storage_def_name=storage_def_name
    ):
        if not encryption_already_enabled_ok:
            raise EncryptionAlreadyEnabledError(
                f"Encryption is already enabled for storage_def={storage_def_name}."
            )
        try:
            credential = atbu_cfg.get_storage_def_encryption_credential(
                storage_def_name=storage_def_name, unlock=True
            )
        except (PasswordAuthenticationFailure, CredentialSecretDerivationError) as ex:
            if ex.message:
                print(ex.message)
            raise

    #
    # Generate/store random 256-bit private key.
    # Associate the user's password with the private key.
    #
    print(
        """
You can require the backup to ask for a password before starting a backup/restore,
or you can allow a backup to proceed automatically without requiring your password.

When you choose the automatic approach which does not require a password, you are
allowing your backup 'private key' to be used automatically by this program. When
doing this, your backup private key is stored in a manner where, not only this
program, but other programs and people who have access to your computer or its
contents may be able to access and use your private key.

You can switch between requiring your password or using the automatic approach as
needed/desired. Regardless of your choice, you should be certain to back up your
security information (i.e., private key, related info) which you can do at any time.

"""
    )
    a = prompt_YN(
        prompt_msg="Choose whether to require password or not.",
        prompt_question="Require a (p)assword or allow (a)utomatic use of your backup's private key? ",
        default_enter_ans="a",
        choices={"p": "[P/a]", "a": "[p/A]"},
    )

    if not credential:
        print(f"Creating key...", end="")
        credential = Credential()
        credential.create_key()
        print(f"created.")

    if debug_mode:
        print(f"set_encryption_password_wizard: key={credential.the_key.hex(' ')}")

    if a == "a":
        cred_operation = CRED_OPERATION_SET_PASSWORD_TO_PRIVATE_KEY
    else:
        cred_operation = CRED_OPERATION_SET_PASSWORD
        print(
            """
You have chosen to require a password before a backup/restore can begin which requires you
to enter a password.
"""
        )
        password = prompt_for_password(
            prompt="Enter a password for this backup:",
            prompt_again="Enter a password for this backup again:",
        )
        credential.prepare_for_new_password()
        credential.set(password=password)
        print(f"Encrypting key...", end="")
        credential.encrypt_key()
        print(f"encrypted.")
        credential.clear_password()

    print(f"Storing...")
    handle_credential_change(
        operation=cred_operation,
        key_to_affect=CRED_KEY_TYPE_ENCRYPTION,
        atbu_cfg=atbu_cfg,
        storage_def_name=storage_def_name,
        credential=credential,
        debug_mode=debug_mode,
    )


def setup_backup_encryption_wizard(
    storage_atbu_cfg: AtbuConfig, storage_def_name: str, debug_mode: bool = False
) -> bool:
    a = prompt_YN(
        prompt_msg="The destination can be encrypted.",
        prompt_question="Would you like encryption enabled?",
    )
    is_encrypted = False
    if a == "y":
        set_encryption_password_wizard(
            atbu_cfg=storage_atbu_cfg,
            storage_def_name=storage_def_name,
            debug_mode=debug_mode,
        )
        print(f"Your key is stored.")
        is_encrypted = True
    else:
        print(f"Files backed up to this destination will *not* be encrypted.")
    print(f"Saving {storage_atbu_cfg.path}")
    storage_atbu_cfg.save_config_file()
    print(f"{storage_atbu_cfg.path} has been saved.")
    return is_encrypted


def is_OAuth2_secret_json_file(path: str) -> bool:
    if path is None or not os.path.isfile(path):
        return False
    # There is a candidate file, see if it ends with .json and contains expected fields...
    logging.info(
        f"Secret seems to reference a file either directly or indirectly: {path}"
    )
    if not path.lower().endswith(".json"):
        logging.warning(f"No .json extension, will not use file as secret: {path}")
        return False
    try:
        with open(path, "r", encoding="utf-8") as cred_file:
            c: dict = json.load(cred_file)
        if c.get("type") is None and c.get("private_key") is None:
            logging.warning(
                f"No 'type' or 'private_key' fields found, will not use file as secret: {path}"
            )
            return False
        # It appears to be a file, such as service account file (i.e., with OAuth2 creds)...
        logging.info(f"Secret will be considered a reference to a file: {path}")
        return True
    except Exception:
        logging.warning(f"Cannot open file, will not use file as secret: {path}")
        return False


def handle_container_auto_creation(
    storage_def_name: str, try_create_if_not_autofind: bool
) -> bool:
    atbu_config = AtbuConfig.access_default_config()
    storage_def_dict = atbu_config.get_storage_def_dict(
        storage_def_name=storage_def_name
    )
    if storage_def_dict is None:
        raise StorageDefinitionNotFoundError(
            f"Cannot find storage defintion '{storage_def_name}'."
        )
    container_name = storage_def_dict[CONFIG_VALUE_NAME_CONTAINER]
    auto_find_name = container_name[-1] == AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR
    if not auto_find_name and not try_create_if_not_autofind:
        # Auto find implies a need to create a container, whereas just
        # the name requires --create-container. If neither auto find
        # nor --create-container, return.
        return False
    if auto_find_name:
        logging.info(
            f"Container name had the {AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR} auto-find/create indicator. "
            f"Searching for unique container name using base name {container_name}..."
        )
    factory = StorageInterfaceFactory.create_factory_from_storage_def_name(
        storage_def_name=storage_def_name
    )
    container_name = factory.storage_def_dict[CONFIG_VALUE_NAME_CONTAINER]
    interface = factory.create_storage_interface()
    try:
        ci = interface.create_container(container_name=container_name)
        logging.info(f"Found/created container name '{ci.name}'.")
        logging.info(f"Updating configuration with that new name.")
    except ContainerAlreadyExistsError:
        logging.warning(
            f"WARNING: The container name chosen already exists: {container_name}"
        )
        logging.warning(
            f"If you do not intend to use this container name, try againg with a different name."
        )
        logging.info(f"Using container name '{container_name}'.")
        logging.info(f"Updating configuration with that existing name.")
    except InvalidContainerNameError:
        logging.error(f"The container name is invalid: {container_name}")
        raise
    except ContainerAutoCreationFailed as ex:
        logging.error(exc_to_string(ex))
        logging.error(
            f"Auto find container with creation failed for name '{container_name}' "
            f"Check the errors and the name."
        )
        raise
    storage_def_dict[CONFIG_VALUE_NAME_CONTAINER] = ci.name
    atbu_config.save_config_file()
    return True


def handle_create_storage_definition(
    storage_def_name: str,
    interface: str,
    provider: str,
    container: str,
    driver_params: str,
    create_container: bool,
    include_iv: bool,
    debug_mode: bool = False,
):
    # Disable MP queued logging to ensure any console
    # logging is in sync with print statements.
    switch_to_non_queued_logging()

    atbu_cfg = AtbuConfig.access_default_config()

    if not storage_def_name:
        raise ValueError(
            f"Invalid name storage_def_name '{storage_def_name}' specified."
        )
    if not interface:
        raise ValueError(f"Invalid name interface '{interface}' specified.")
    if not provider:
        raise ValueError(f"Invalid name provider '{provider}' specified.")
    if not container:
        raise ValueError(f"Invalid name container '{container}' specified.")
    if not driver_params:
        raise ValueError(f"Invalid name container '{driver_params}' specified.")

    if atbu_cfg.is_storage_def_exists(storage_def_name=storage_def_name):
        raise StorageDefinitionAlreadyExists(
            f"The storage definition '{storage_def_name}' already exists. Delete it first, or choose a different name."
        )

    storage_def_name = storage_def_name.lower()
    if not is_storage_def_name_ok(storage_def_name=storage_def_name):
        raise InvalidStorageDefinitionName(
            f"The name '{storage_def_name}' is invalid. Allowed characters: lowercase alphanumeric, underscore, hyphen/dash."
        )

    secret = CredentialByteArray()
    m = re.match(r".*,secret=([^,]+)", driver_params)
    if m:
        # A base64 secret may contain '=' so remove and parse separately.
        secret = CredentialByteArray(m.groups()[0].encode("utf-8"))
        driver_params = re.sub(r",\s*secret=([^,]+)", "", driver_params)
    other_driver_kv_pairs = dict(
        kv_pair.split("=") for kv_pair in driver_params.split(",")
    )

    other_driver_kv_pairs[CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET] = secret

    _, _ = atbu_cfg.create_storage_def(
        interface_type=interface,
        provider_id=provider,
        container=container,
        other_driver_kv_pairs=other_driver_kv_pairs,
        unique_storage_def_name=storage_def_name,
        include_iv=include_iv,
    )

    operation = CRED_OPERATION_SET_PASSWORD
    secret_str = secret.decode("utf-8")
    if os.path.isfile(secret_str) and is_OAuth2_secret_json_file(path=secret_str):
        # Yes, secret is directly an existing file name.
        operation = CRED_OPERATION_SET_PASSWORD_TO_FILENAME
    else:
        # Not a direct file name.
        # Check if env var.
        ev = os.getenv(secret_str)
        if isinstance(ev, str) and is_OAuth2_secret_json_file(path=ev):
            # Yes, env var is file.
            operation = CRED_OPERATION_SET_PASSWORD_TO_ENVVAR

    try:
        # Save to keyring and indirect storage definition to reference the keyring creds.
        handle_credential_change(
            operation=operation,
            key_to_affect=CRED_KEY_TYPE_STORAGE,
            atbu_cfg=atbu_cfg,
            storage_def_name=storage_def_name,
            credential=secret,
            debug_mode=debug_mode,
        )

        atbu_cfg.save_config_file()

        logging.info(f"Storage definition {storage_def_name} saved.")
        logging.debug(f"Storage definition {storage_def_name} saved to {atbu_cfg.path}")

        is_encrypted = setup_backup_encryption_wizard(
            storage_atbu_cfg=atbu_cfg,
            storage_def_name=storage_def_name,
            debug_mode=debug_mode,
        )

        if not is_encrypted:
            logging.warning(
                f"WARNING: The storage definition '{storage_def_name}' will *not* be encrypted."
            )
        else:
            logging.info(
                f"The storage definition '{storage_def_name}' will be encrypted."
            )

        handle_container_auto_creation(
            storage_def_name=storage_def_name,
            try_create_if_not_autofind=create_container,
        )

        logging.info(f"Storage definition {storage_def_name} successfully created.")
    except Exception as ex:
        message = str(ex)
        if isinstance(ex, AtbuException):
            message = ex.message
        if atbu_cfg.is_storage_def_exists(storage_def_name=storage_def_name):
            logging.warning(
                f"Due to error, deleting storage definition '{storage_def_name}'. {message} {exc_to_string(ex)}"
            )
            atbu_cfg.delete_storage_def(storage_def_name=storage_def_name)
            atbu_cfg.save_config_file()
        raise


def handle_delete_storage_definition(
    storage_def_name: str,
    skip_confirmation_prompt: bool,
    delete_backup_info: bool,
):
    """Delete a storage definition.

    Args:
        storage_def_name: The storage definition name.
        skip_confirmation_prompt: If True, then no
            prompting will occur, definition will be
            deleted.
        delete_backup_info: If True, the backup information
            will also be deleted, if False the backup
            information will not be deleted. If None,
            and skip_confirmation_prompt==False, the
            user will be prompted about backup information,
            otherwise in the case of not prompting, the
            backup information will not be deleted. The
            goal is to be cautious about deleting backup
            informoation despite it being recoverable from
            a good backup. Essentially, the user must be
            explicitly one way or another about including
            backup information in the deletion process.
    """
    switch_to_non_queued_logging()
    bin_name = get_trash_bin_name()
    atbu_cfg = AtbuConfig.access_default_config()
    if not atbu_cfg.is_storage_def_exists(storage_def_name=storage_def_name):
        print(f"The storage definition '{storage_def_name}' does not exist.")
        return

    existing_backup_info = atbu_cfg.get_backup_info_file_paths(
        storage_def_name=storage_def_name
    )

    if not skip_confirmation_prompt:
        print(
            f"""*** WARNING *** WARNING *** WARNING *** WARNING *** WARNING *** WARNING
The storage definition '{storage_def_name}' exists.
If this is an encrypted backup where the private key is not backed up,
you *will* *lose* *access* to all data in this backup if you delete this
configuration.
"""
        )
        if (
            delete_backup_info is not None  # If user was explicit
            and delete_backup_info  # and user said Yes to deleting BI
            and len(existing_backup_info) > 0  # and there is BI to delete
        ):
            print()
            print(
                f"""If you delete this storage definition, the following backup information
will also be deleted:
"""
            )
            print()
            for bip in existing_backup_info:
                print(f"    {bip}")
            print()
        a = prompt_YN(
            prompt_msg=f"You are about to delete a backup storage definition.",
            prompt_question=f"Are you certain you want to delete '{storage_def_name}' ",
            default_enter_ans="n",
        )
        if a != "y":
            print("The storage definition will not be deleted.")
            return
        if (
            delete_backup_info is None  # user was not explicit
            and len(existing_backup_info) > 0  # there is BI to delete
        ):
            print()
            print(f"This storage definition also has backup information as follows:")
            print()
            for bip in existing_backup_info:
                print(f"    {bip}")
            print()
            print(
                f"""If you are planning to immediately re-import this same configuration,
or you otherwise need this backup information, you should not delete it.

While this backup information is typically recoverable from a non-corrupt
backup, it is generally a good idea to err on the side of caution and back
it up if you are uncertain.

If you choose 'y' below to delete the above files, they will no longer be
available. This app will attempt to send the files to the {bin_name},
so you may be able to recover them from there if needed.

If you choose *not* to delete these files along with the storage definition,
and you later attempt import/recovery of a storage definition of the same name,
the lingering presence of these older files may interfere or be out of sync with
that newly imported configuration.
                """
            )
            a = prompt_YN(
                prompt_msg=f"Choose whether to also delete the above backup information files (a=abort).",
                prompt_question=f"Delete backup information files along with '{storage_def_name}' ",
                default_enter_ans="a",
                choices={"y": "[Y/n/a]", "n": "[y/N/a]", "a": "[y/n/A]"},
            )
            if a == "a":
                print("The storage definition will not be deleted.")
                return
            delete_backup_info = a == "y"
    logging.info(f"Deleting storage definition '{storage_def_name}'...")
    atbu_cfg.delete_storage_def(storage_def_name=storage_def_name)
    logging.info(f"Saving configuration file {atbu_cfg.path}...")
    atbu_cfg.save_config_file()
    if delete_backup_info:
        logging.info("Deleting backup information...")
        for bip in existing_backup_info:
            logging.info(f"    Sending file to {bin_name}: {bip}")
            send2trash(bip)
    elif len(existing_backup_info) > 0:
        logging.info(f"Skipping deletion of the following backup information...")
        for bip in existing_backup_info:
            logging.info(f"    {bip}")
    else:
        logging.info(f"There is no backup information for '{storage_def_name}'.")
    logging.info(f"The storage definition '{storage_def_name}' was deleted.")


def handle_creds(args):
    logging.debug(f"handle_creds")
    storage_def_name = args.storage_def
    orig_storage_def_name = storage_def_name  # pylint: disable=unused-variable
    storage_atbu_cfg: AtbuConfig
    show_secrets = False
    if hasattr(args, "show_secrets"):
        show_secrets = args.show_secrets
    if not is_existing_filesystem_storage_path(storage_location=storage_def_name):
        #
        # Storage definition is not a filesystem storage.
        # It must therefore be a storage-def-specifier or simply/directly the storage_def_name itself.
        # Resolve to a storage_def_name...
        #
        parsed_name = parse_storage_def_specifier(storage_location=storage_def_name)
        if parsed_name:
            storage_def_name = parsed_name
        storage_atbu_cfg = AtbuConfig.access_default_config()
    else:
        #
        # Filesystem storage.
        #
        storage_atbu_cfg, storage_def_name, _ = AtbuConfig.access_filesystem_config(
            storage_location_path=storage_def_name,
            create_if_not_exist=True,
            prompt_to_create=True,
        )
    if args.subcmd == "create-storage-def":
        handle_create_storage_definition(
            storage_def_name=storage_def_name,
            interface=args.interface,
            provider=args.provider,
            container=args.container,
            driver_params=args.driver_params,
            create_container=args.create_container,
            include_iv=args.include_iv,
            debug_mode=show_secrets,
        )
    elif args.subcmd == "delete-storage-def":
        handle_delete_storage_definition(
            storage_def_name=storage_def_name,
            skip_confirmation_prompt=args.force,
            delete_backup_info=args.delete_backup_info,
        )
    elif args.subcmd == "export":
        backup_file_path = args.filename
        if not backup_file_path:
            raise InvalidCommandLineArgument(
                f"You must specify a filename to save the storage definition to."
            )
        if os.path.isdir(backup_file_path):
            raise InvalidCommandLineArgument(
                f"The specified file is a directory: {backup_file_path}"
            )
        if os.path.isfile(backup_file_path) and not args.overwrite:
            raise InvalidCommandLineArgument(
                f"The file already exists: {backup_file_path}"
            )
        storage_atbu_cfg.save_storage_def_with_cleartext_secrets(
            storage_def_name=storage_def_name, backup_file_path=backup_file_path
        )
    elif args.subcmd == "import":
        switch_to_non_queued_logging()
        # TODO: Handle args.create_config
        backup_file_path = args.filename
        if not backup_file_path:
            raise InvalidCommandLineArgument(
                f"You must specify the filename of the storage definition and "
                f"credentials to restore."
            )
        if os.path.isdir(backup_file_path):
            raise InvalidCommandLineArgument(
                f"The specified file is a directory: {backup_file_path}"
            )
        if not os.path.isfile(backup_file_path):
            raise InvalidCommandLineArgument(
                f"The file does not exist: {backup_file_path}"
            )
        storage_atbu_cfg.restore_storage_def(
            storage_def_new_name=storage_def_name,
            backup_file_path=backup_file_path,
            prompt_if_exists=args.prompt,
        )
    elif args.subcmd == "set-password":
        operation = CRED_OPERATION_SET_PASSWORD
        if args.key_type == CRED_KEY_TYPE_ENCRYPTION:
            if args.password is not None:
                logging.error(
                    f"A command line password cannot be specified for {CRED_KEY_TYPE_ENCRYPTION}, "
                    f"only for {CRED_KEY_TYPE_STORAGE}."
                )
                logging.info(
                    f"To change the {CRED_KEY_TYPE_ENCRYPTION} password, do not specify a password "
                    f"on the command line which allows a console UI to prompt you."
                )
                raise CredentialRequestInvalidError(
                    f"A command line password cannot be specified for {CRED_KEY_TYPE_ENCRYPTION}."
                )
            set_encryption_password_wizard(
                atbu_cfg=storage_atbu_cfg,
                storage_def_name=storage_def_name,
                encryption_already_enabled_ok=True,
                debug_mode=show_secrets,
            )
        elif args.key_type == CRED_KEY_TYPE_STORAGE:
            if args.password is None:
                raise CredentialRequestInvalidError(
                    f"A password is required to set the password for the {CRED_KEY_TYPE_STORAGE}."
                )
            handle_credential_change(
                operation=operation,
                key_to_affect=args.key_type,
                atbu_cfg=storage_atbu_cfg,
                storage_def_name=storage_def_name,
                credential=args.password,
                debug_mode=show_secrets,
            )
        else:
            raise InvalidStateError(f"Unexpected key type '{args.key_type}'")
    elif args.subcmd == "set-password-filename":
        if args.filename is None:
            raise CredentialRequestInvalidError(
                f"A filename is required to set the password filename for the {CRED_KEY_TYPE_STORAGE}."
            )
        operation = CRED_OPERATION_SET_PASSWORD_TO_FILENAME
        handle_credential_change(
            operation=operation,
            key_to_affect=CRED_KEY_TYPE_STORAGE,
            atbu_cfg=storage_atbu_cfg,
            storage_def_name=storage_def_name,
            credential=args.filename,
            debug_mode=show_secrets,
        )
    elif args.subcmd == "set-password-envvar":
        operation = CRED_OPERATION_SET_PASSWORD_TO_ENVVAR
        if args.key_type == CRED_KEY_TYPE_STORAGE:
            handle_credential_change(
                operation=operation,
                key_to_affect=args.key_type,
                atbu_cfg=storage_atbu_cfg,
                storage_def_name=storage_def_name,
                credential=args.env_var,
                debug_mode=show_secrets,
            )
        else:
            raise InvalidStateError(f"Key type '{args.key_type} not allowed.'")
    else:
        raise InvalidCommandLineArgument("Unknown command.")
